import datetime as dt

import pytest
import pandas as pd
import numpy as np
from numpy.testing import assert_array_equal
from pandas.testing import assert_series_equal
import shapely.geometry as sg

from .context import gtfs_kit, cairns, cairns_dates, cairns_trip_stats
from gtfs_kit import helpers as gkh


def test_timestr_to_seconds():
    timestr1 = "01:01:01"
    seconds1 = 3600 + 60 + 1
    timestr2 = "25:01:01"
    seconds2 = 25 * 3600 + 60 + 1
    assert gkh.timestr_to_seconds(timestr1) == seconds1
    assert gkh.timestr_to_seconds(seconds1, inverse=True) == timestr1
    assert gkh.timestr_to_seconds(seconds2, inverse=True) == timestr2
    assert gkh.timestr_to_seconds(timestr2, mod24=True) == seconds1
    assert gkh.timestr_to_seconds(seconds2, mod24=True, inverse=True) == timestr1
    # Test error handling
    assert np.isnan(gkh.timestr_to_seconds(seconds1))
    assert np.isnan(gkh.timestr_to_seconds(timestr1, inverse=True))


def test_timestr_mod24():
    timestr1 = "01:01:01"
    assert gkh.timestr_mod24(timestr1) == timestr1
    timestr2 = "25:01:01"
    assert gkh.timestr_mod24(timestr2) == timestr1


def test_datestr_to_date():
    datestr = "20140102"
    date = dt.date(2014, 1, 2)
    assert gkh.datestr_to_date(datestr) == date
    assert gkh.datestr_to_date(date, inverse=True) == datestr


def test_is_metric():
    assert gkh.is_metric("m")
    assert gkh.is_metric("km")
    assert not gkh.is_metric("ft")
    assert not gkh.is_metric("mi")
    assert not gkh.is_metric("bingo")


def test_get_convert_dist():
    di = "mi"
    do = "km"
    f = gkh.get_convert_dist(di, do)
    assert f(1) == 1.609_344


def test_get_segment_length():
    s = sg.LineString([(0, 0), (1, 0)])
    p = sg.Point((1 / 2, 0))
    assert gkh.get_segment_length(s, p) == 1 / 2
    q = sg.Point((1 / 3, 0))
    assert gkh.get_segment_length(s, p, q) == pytest.approx(1 / 6)
    p = sg.Point((0, 1 / 2))
    assert gkh.get_segment_length(s, p) == 0


def test_get_max_runs():
    x = [7, 1, 2, 7, 7, 1, 2]
    get = gkh.get_max_runs(x)
    expect = np.array([[0, 1], [3, 5]])
    assert_array_equal(get, expect)


def test_get_peak_indices():
    times = [0, 10, 20, 30, 31, 32, 40]
    counts = [7, 1, 2, 7, 7, 1, 2]
    get = gkh.get_peak_indices(times, counts)
    expect = [0, 1]
    assert_array_equal(get, expect)

    counts = [0, 1, 0]
    times = [18000, 21600, 28800]
    get = gkh.get_peak_indices(times, counts)
    expect = [1, 2]
    assert_array_equal(get, expect)

    counts = [0, 0, 0]
    times = [18000, 21600, 28800]
    get = gkh.get_peak_indices(times, counts)
    expect = [0, 3]
    assert_array_equal(get, expect)


def test_almost_equal():
    f = pd.DataFrame([[1, 2], [3, 4]], columns=["a", "b"])
    assert gkh.almost_equal(f, f)
    g = pd.DataFrame([[4, 3], [2, 1]], columns=["b", "a"])
    assert gkh.almost_equal(f, g)
    h = pd.DataFrame([[1, 2], [5, 4]], columns=["a", "b"])
    assert not gkh.almost_equal(f, h)
    h = pd.DataFrame()
    assert not gkh.almost_equal(f, h)


def test_is_not_null():
    f = None
    c = "foo"
    assert not gkh.is_not_null(f, c)

    f = pd.DataFrame(columns=["bar", c])
    assert not gkh.is_not_null(f, c)

    f = pd.DataFrame([[1, np.nan]], columns=["bar", c])
    assert not gkh.is_not_null(f, c)

    f = pd.DataFrame([[1, np.nan], [2, 2]], columns=["bar", c])
    assert gkh.is_not_null(f, c)


def test_get_active_trips_df():
    f = pd.DataFrame({"start_time": [1, 2, 3, 4, 5], "end_time": [6, 7, 8, 9, 10]})
    expect = pd.Series(
        index=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10], data=[1, 2, 3, 4, 5, 4, 3, 2, 1, 0]
    )
    get = gkh.get_active_trips_df(f)
    assert_series_equal(get, expect)

    f = pd.DataFrame({"start_time": [1, 2, 3, 4, 5], "end_time": [2, 4, 6, 8, 10]})
    expect = pd.Series(index=[1, 2, 3, 4, 5, 6, 8, 10], data=[1, 1, 2, 2, 3, 2, 1, 0])
    get = gkh.get_active_trips_df(f)
    assert_series_equal(get, expect)


def test_downsample():
    ts = cairns.compute_route_time_series(cairns_trip_stats, cairns_dates, freq="6h")
    f = gkh.downsample(ts, "6h")
    assert ts.equals(f)

    f = gkh.downsample(ts, "12h")
    assert f.shape[0] == ts.shape[0] / 2
    assert pd.tseries.frequencies.to_offset(f.index.freq) == "12h"


def test_unstack_time_series():
    dates = cairns_dates
    for split_directions in [True, False]:
        f = cairns.compute_stop_time_series(
            dates, freq="12h", split_directions=split_directions
        )
        g = gkh.unstack_time_series(f)
        expect_cols = {"datetime", "indicator", "value", "stop_id"}
        if split_directions:
            expect_cols.add("direction_id")

        assert set(g.columns) == expect_cols
        assert g.shape[0] == f.shape[0] * f.shape[1]


def test_restack_time_series():
    dates = cairns_dates
    for split_directions in [True, False]:
        f = cairns.compute_stop_time_series(
            dates, freq="12h", split_directions=split_directions
        )
        g = gkh.restack_time_series(gkh.unstack_time_series(f))
        assert set(g.columns) == set(f.columns)
        assert f.shape[0] == g.shape[0]


def test_longest_subsequence():
    dates = [
        ("2015-02-03", "name1"),
        ("2015-02-04", "nameg"),
        ("2015-02-04", "name5"),
        ("2015-02-05", "nameh"),
        ("1929-03-12", "name4"),
        ("2023-07-01", "name7"),
        ("2015-02-07", "name0"),
        ("2015-02-08", "nameh"),
        ("2015-02-15", "namex"),
        ("2015-02-09", "namew"),
        ("1980-12-23", "name2"),
        ("2015-02-12", "namen"),
        ("2015-02-13", "named"),
    ]

    assert gkh.longest_subsequence(dates, "weak") == [
        ("2015-02-03", "name1"),
        ("2015-02-04", "name5"),
        ("2015-02-05", "nameh"),
        ("2015-02-07", "name0"),
        ("2015-02-08", "nameh"),
        ("2015-02-09", "namew"),
        ("2015-02-12", "namen"),
        ("2015-02-13", "named"),
    ]

    from operator import itemgetter

    assert gkh.longest_subsequence(dates, "weak", key=itemgetter(0)) == [
        ("2015-02-03", "name1"),
        ("2015-02-04", "nameg"),
        ("2015-02-04", "name5"),
        ("2015-02-05", "nameh"),
        ("2015-02-07", "name0"),
        ("2015-02-08", "nameh"),
        ("2015-02-09", "namew"),
        ("2015-02-12", "namen"),
        ("2015-02-13", "named"),
    ]

    indices = set(gkh.longest_subsequence(dates, key=itemgetter(0), index=True))
    assert [e for i, e in enumerate(dates) if i not in indices] == [
        ("2015-02-04", "nameg"),
        ("1929-03-12", "name4"),
        ("2023-07-01", "name7"),
        ("2015-02-15", "namex"),
        ("1980-12-23", "name2"),
    ]


def test_make_ids():
    assert gkh.make_ids(10, "s") == [
        "s0",
        "s1",
        "s2",
        "s3",
        "s4",
        "s5",
        "s6",
        "s7",
        "s8",
        "s9",
    ]
    assert gkh.make_ids(11, "s") == [
        "s00",
        "s01",
        "s02",
        "s03",
        "s04",
        "s05",
        "s06",
        "s07",
        "s08",
        "s09",
        "s10",
    ]



import pytest
import pandas as pd
from pandas.testing import assert_series_equal
import numpy as np
import shapely.geometry as sg
import geopandas as gp

from .context import gtfs_kit, DATA_DIR, sample, cairns, cairns_dates, cairns_trip_stats
from gtfs_kit import miscellany as gkm
from gtfs_kit import shapes as gks


def test_summarize():
    feed = sample.copy()

    with pytest.raises(ValueError):
        gkm.summarize(feed, "bad_table")

    for table in [None, "stops"]:
        f = gkm.summarize(feed, table)
        assert isinstance(f, pd.DataFrame)
        expect_cols = {
            "table",
            "column",
            "num_values",
            "num_nonnull_values",
            "num_unique_values",
            "min_value",
            "max_value",
        }
        assert set(f.columns) == expect_cols
        assert f.shape[0]


def test_describe():
    feed = sample.copy()  # No distances here
    a = gkm.describe(feed)
    assert isinstance(a, pd.DataFrame)
    assert set(a.columns) == set(["indicator", "value"])


def test_assess_quality():
    feed = sample.copy()  # No distances here
    a = gkm.assess_quality(feed)
    assert isinstance(a, pd.DataFrame)
    assert set(a.columns) == set(["indicator", "value"])


def test_convert_dist():
    # Test with no distances
    feed1 = cairns.copy()  # No distances here
    feed2 = gkm.convert_dist(feed1, "mi")
    assert feed2.dist_units == "mi"

    # Test with distances and identity conversion
    feed1 = gks.append_dist_to_shapes(feed1)
    feed2 = gkm.convert_dist(feed1, feed1.dist_units)
    assert feed1 == feed2

    # Test with proper conversion
    feed2 = gkm.convert_dist(feed1, "m")
    assert_series_equal(
        feed2.shapes["shape_dist_traveled"] / 1000, feed1.shapes["shape_dist_traveled"]
    )


def test_compute_feed_stats_0():
    feed = cairns.copy()
    trip_stats = cairns_trip_stats
    feed.routes.route_type.iat[0] = 2  # Another route type besides 3
    for split_route_types in [True, False]:
        f = gkm.compute_feed_stats_0(
            feed, trip_stats, split_route_types=split_route_types
        )
        # Should be a data frame
        assert isinstance(f, pd.core.frame.DataFrame)
        # Should contain the correct columns
        expect_cols = {
            "num_trips",
            "num_trip_starts",
            "num_trip_ends",
            "num_routes",
            "num_stops",
            "peak_num_trips",
            "peak_start_time",
            "peak_end_time",
            "service_duration",
            "service_distance",
            "service_speed",
        }
        if split_route_types:
            expect_cols.add("route_type")

        assert set(f.columns) == expect_cols


def test_compute_feed_stats():
    feed = cairns.copy()
    dates = cairns_dates + ["20010101"]
    trip_stats = cairns_trip_stats
    feed.routes.route_type.iat[0] = 2  # Another route type besides 3
    for split_route_types in [True, False]:
        f = gkm.compute_feed_stats(
            feed, trip_stats, dates, split_route_types=split_route_types
        )
        # Should be a data frame
        assert isinstance(f, pd.core.frame.DataFrame)
        # Should have the correct dates
        assert f.date.tolist() == cairns_dates
        # Should contain the correct columns
        expect_cols = {
            "num_trips",
            "num_trip_starts",
            "num_trip_ends",
            "num_routes",
            "num_stops",
            "peak_num_trips",
            "peak_start_time",
            "peak_end_time",
            "service_duration",
            "service_distance",
            "service_speed",
            "date",
        }
        if split_route_types:
            expect_cols.add("route_type")

        assert set(f.columns) == expect_cols

        # Empty dates should yield empty DataFrame
        f = gkm.compute_feed_stats(
            feed, trip_stats, [], split_route_types=split_route_types
        )
        assert f.empty


def test_compute_feed_time_series():
    feed = cairns.copy()
    feed.routes.route_type.iat[0] = 2  # Add another route type
    dates = cairns_dates + ["20010101"]
    trip_stats = cairns_trip_stats

    for split_route_types in [True, False]:
        f = gkm.compute_feed_time_series(
            feed, trip_stats, dates, freq="12h", split_route_types=split_route_types
        )

        # Should have correct column level names
        if split_route_types:
            assert set(f.columns.names) == {"indicator", "route_type"}
        else:
            assert set(f.columns.names) == {"indicator"}

        # Should have the correct level 0 columns
        expect_cols = {
            "num_trip_starts",
            "num_trip_ends",
            "num_trips",
            "service_distance",
            "service_duration",
            "service_speed",
        }
        if split_route_types:
            assert set(f.columns.levels[0]) == expect_cols
        else:
            assert set(f.columns) == expect_cols

        # Should have correct index names
        assert f.index.name == "datetime"

        # Should have the correct number of rows: 2 (for the 12H freq) times
        # 3 (for the three-date span)
        assert f.shape[0] == 2 * 3

        # Empty check
        f = gkm.compute_feed_time_series(
            feed, trip_stats, [], split_route_types=split_route_types
        )
        assert f.empty


def test_create_shapes():
    feed1 = cairns.copy()
    # Remove a trip shape
    trip_id = "CNS2014-CNS_MUL-Weekday-00-4165878"
    feed1.trips.loc[feed1.trips["trip_id"] == trip_id, "shape_id"] = np.nan
    feed2 = gkm.create_shapes(feed1)
    # Should create only 1 new shape
    assert len(set(feed2.shapes["shape_id"]) - set(feed1.shapes["shape_id"])) == 1

    feed2 = gkm.create_shapes(feed1, all_trips=True)
    # Number of shapes should equal number of unique stop sequences
    st = feed1.stop_times.sort_values(["trip_id", "stop_sequence"])
    stop_seqs = set(
        [tuple(group["stop_id"].values) for __, group in st.groupby("trip_id")]
    )
    assert feed2.shapes["shape_id"].nunique() == len(stop_seqs)


def test_compute_bounds():
    feed = cairns.copy()
    minlon, minlat, maxlon, maxlat = gkm.compute_bounds(feed)
    # Bounds should be in the ball park
    assert 145 < minlon < 146
    assert 145 < maxlon < 146
    assert -18 < minlat < -15
    assert -18 < maxlat < -15

    # A one-stop bounds should be the stop
    stop_id = feed.stops.stop_id.iat[0]
    minlon, minlat, maxlon, maxlat = gkm.compute_bounds(feed, [stop_id])
    expect_lon, expect_lat = feed.stops.loc[
        lambda x: x.stop_id == stop_id, ["stop_lon", "stop_lat"]
    ].values[0]
    assert minlon == expect_lon
    assert minlat == expect_lat
    assert maxlon == expect_lon
    assert maxlat == expect_lat


def test_compute_convex_hull():
    feed = cairns.copy()
    hull = gkm.compute_convex_hull(feed)
    assert isinstance(hull, sg.Polygon)
    # Hull should encompass all stops
    m = sg.MultiPoint(feed.stops[["stop_lon", "stop_lat"]].values)
    assert hull.contains(m)

    # A one-stop hull should be the stop
    stop_id = feed.stops.stop_id.iat[0]
    hull = gkm.compute_convex_hull(feed, [stop_id])
    lon, lat = np.array(hull.coords)[0]
    expect_lon, expect_lat = feed.stops.loc[
        lambda x: x.stop_id == stop_id, ["stop_lon", "stop_lat"]
    ].values[0]
    assert lon == expect_lon
    assert lat == expect_lat


def test_compute_centroid():
    feed = cairns.copy()
    centroid = gkm.compute_centroid(feed)
    assert isinstance(centroid, sg.Point)
    # Centroid should lie within bounds
    lon, lat = centroid.coords[0]
    bounds = gkm.compute_bounds(feed)
    assert bounds[0] < lon < bounds[2]
    assert bounds[1] < lat < bounds[3]

    # A one-stop centroid should be the stop
    stop_id = feed.stops.stop_id.iat[0]
    centroid = gkm.compute_centroid(feed, [stop_id])
    lon, lat = centroid.coords[0]
    expect_lon, expect_lat = feed.stops.loc[
        lambda x: x.stop_id == stop_id, ["stop_lon", "stop_lat"]
    ].values[0]
    assert lon == expect_lon
    assert lat == expect_lat


def test_restrict_to_dates():
    feed1 = cairns.copy()
    dates = feed1.get_first_week()[6:]
    feed2 = gkm.restrict_to_dates(feed1, dates)
    # Should have correct agency
    assert feed2.agency.equals(feed1.agency)
    # Should have correct dates
    assert set(feed2.get_dates()) < set(feed1.get_dates())
    # Should have correct trips
    assert set(feed2.trips.trip_id) < set(feed1.trips.trip_id)
    # Should have correct routes
    assert set(feed2.routes.route_id) < set(feed1.routes.route_id)
    # Should have correct shapes
    assert set(feed2.shapes.shape_id) < set(feed1.shapes.shape_id)
    # Should have correct stops
    assert set(feed2.stops.stop_id) < set(feed1.stops.stop_id)

    # Try again with date out of range
    dates = ["20180101"]
    feed2 = gkm.restrict_to_dates(feed1, dates)
    assert feed2.agency.equals(feed1.agency)
    assert feed2.trips.empty
    assert feed2.routes.empty
    assert feed2.shapes.empty
    assert feed2.stops.empty
    assert feed2.stop_times.empty


def test_restrict_to_routes():
    feed1 = cairns.copy()
    route_ids = feed1.routes["route_id"][:2].tolist()
    feed2 = gkm.restrict_to_routes(feed1, route_ids)
    # Should have correct routes
    assert set(feed2.routes["route_id"]) == set(route_ids)
    # Should have correct trips
    trip_ids = feed1.trips[feed1.trips["route_id"].isin(route_ids)]["trip_id"]
    assert set(feed2.trips["trip_id"]) == set(trip_ids)
    # Should have correct shapes
    shape_ids = feed1.trips[feed1.trips["trip_id"].isin(trip_ids)]["shape_id"]
    assert set(feed2.shapes["shape_id"]) == set(shape_ids)
    # Should have correct stops
    stop_ids = feed1.stop_times[feed1.stop_times["trip_id"].isin(trip_ids)]["stop_id"]
    assert set(feed2.stop_times["stop_id"]) == set(stop_ids)


def test_restrict_to_area():
    feed1 = cairns.copy()
    area = gp.read_file(DATA_DIR / "cairns_square_stop_750070.geojson")
    feed2 = gkm.restrict_to_area(feed1, area)
    # Should have correct routes
    rsns = ["120", "120N"]
    assert set(feed2.routes["route_short_name"]) == set(rsns)
    # Should have correct trips
    route_ids = feed1.routes[feed1.routes["route_short_name"].isin(rsns)]["route_id"]
    trip_ids = feed1.trips[feed1.trips["route_id"].isin(route_ids)]["trip_id"]
    assert set(feed2.trips["trip_id"]) == set(trip_ids)
    # Should have correct shapes
    shape_ids = feed1.trips[feed1.trips["trip_id"].isin(trip_ids)]["shape_id"]
    assert set(feed2.shapes["shape_id"]) == set(shape_ids)
    # Should have correct stops
    stop_ids = feed1.stop_times[feed1.stop_times["trip_id"].isin(trip_ids)]["stop_id"]
    assert set(feed2.stop_times["stop_id"]) == set(stop_ids)


@pytest.mark.slow
def test_compute_screen_line_counts():
    feed = cairns.append_dist_to_stop_times()
    dates = cairns_dates + ["20010101"]

    # Load screen line
    path = DATA_DIR / "cairns_screen_lines.geojson"
    screen_lines = gp.read_file(path)
    f = gkm.compute_screen_line_counts(feed, screen_lines, dates)

    # Should have correct columns
    expect_cols = {
        "date",
        "trip_id",
        "route_id",
        "route_short_name",
        "shape_id",
        "screen_line_id",
        "crossing_distance",
        "crossing_time",
        "crossing_direction",
    }
    assert set(f.columns) == expect_cols

    # Should have both directions
    assert set(f.crossing_direction.unique()) == {-1, 1}

    # Should only have feed dates
    assert f.date.unique().tolist() == cairns_dates

    # Empty check
    f = gkm.compute_screen_line_counts(feed, screen_lines, [])
    assert f.empty



"""
Functions about stop times.
"""

from __future__ import annotations
from typing import Iterable, TYPE_CHECKING
import json

import pandas as pd
import numpy as np

from . import helpers as hp

# Help mypy but avoid circular imports
if TYPE_CHECKING:
    from .feed import Feed


def get_stop_times(feed: "Feed", date: str | None = None) -> pd.DataFrame:
    """
    Return ``feed.stop_times``.
    If a date (YYYYMMDD date string) is given, then subset the result to only those
    stop times with trips active on the date.
    """
    if date is None:
        f = feed.stop_times.copy()
    else:
        trip_ids = feed.get_trips(date).trip_id
        f = feed.stop_times.loc[lambda x: x.trip_id.isin(trip_ids)].copy()

    return f


def append_dist_to_stop_times(feed: "Feed") -> "Feed":
    """
    Calculate and append the optional ``shape_dist_traveled`` column in
    ``feed.stop_times`` in terms of the distance units ``feed.dist_units``.
    Return the resulting Feed.

    This does not always give accurate results.
    The algorithm works as follows.
    Compute the ``shape_dist_traveled`` field by using Shapely to
    measure the distance of a stop along its trip LineString.
    If for a given trip this process produces a non-monotonically
    increasing, hence incorrect, list of (cumulative) distances, then
    fall back to estimating the distances as follows.

    Set the first distance to 0, the last to the length of the trip shape,
    and leave the remaining ones computed above.
    Choose the longest increasing subsequence of that new set of
    distances and use them and their corresponding departure times to linearly
    interpolate the rest of the distances.
    """
    # Get stop and shape geometries as dictionaries
    geom_by_stop = feed.build_geometry_by_stop(use_utm=True)
    geom_by_shape = feed.build_geometry_by_shape(use_utm=True)

    # Memoize distance by stop by shape to avoid repeating calculations
    dist_by_stop_by_shape = {shape: {} for shape in geom_by_shape}

    def compute_dist(group):
        g = group.copy()

        # Compute the distances of the stops along this trip and memoize.
        shape = g.shape_id.iat[0]
        linestring = geom_by_shape[shape]
        dists = []
        for stop in g.stop_id.values:
            if stop in dist_by_stop_by_shape[shape]:
                d = dist_by_stop_by_shape[shape][stop]
            else:
                d = linestring.project(geom_by_stop[stop])
                dist_by_stop_by_shape[shape][stop] = d
            dists.append(d)

        s = sorted(dists)
        D = linestring.length
        dists_are_reasonable = all([d < D + 100 for d in dists])

        if dists_are_reasonable and s == dists:
            # Good
            g["shape_dist_traveled"] = dists
        elif dists_are_reasonable and s == dists[::-1]:
            # Good after reversal.
            # This happens when the direction of the linestring
            # opposes the direction of the vehicle trip.
            dists = dists[::-1]
            g["shape_dist_traveled"] = dists
        else:
            # Bad. Redo using interpolation on a good subset of dists.
            dists = np.array([0] + dists[1:-1] + [D])
            ix = hp.longest_subsequence(dists, index=True)
            good_dists = np.take(dists, ix)
            g["shape_dist_traveled"] = np.interp(
                g["departure_time_s"], g.iloc[ix]["departure_time_s"], good_dists
            )

            # Update dist dictionary with new and improved dists
            for row in g[["stop_id", "shape_dist_traveled"]].itertuples(index=False):
                dist_by_stop_by_shape[shape][row.stop_id] = row.shape_dist_traveled

        return g

    cols = [c for c in feed.stop_times.columns if c != "shape_dist_traveled"]
    new_cols = cols + ["shape_dist_traveled"]

    m_to_dist = hp.get_convert_dist("m", feed.dist_units)
    st = (
        feed.stop_times.filter(cols)
        .merge(feed.trips.filter(["trip_id", "shape_id"]))
        # Convert departure times to seconds to ease calculatios
        .assign(departure_time_s=lambda x: x.departure_time.map(hp.timestr_to_seconds))
        .sort_values(["trip_id", "stop_sequence"])
        .groupby("trip_id", group_keys=False)
        .apply(compute_dist)
        .reset_index()
        # Convert distances from meters to feed's distance units
        .assign(
            shape_dist_traveled=lambda x: x.shape_dist_traveled.map(
                m_to_dist, na_action="ignore"
            )
        )
        .filter(new_cols)
    )

    # Create new feed
    new_feed = feed.copy()
    new_feed.stop_times = st

    return new_feed


def get_start_and_end_times(feed: "Feed", date: str | None = None) -> list[str]:
    """
    Return the first departure time and last arrival time
    (HH:MM:SS time strings) listed in ``feed.stop_times``, respectively.
    Restrict to the given date (YYYYMMDD string) if specified.
    """
    st = feed.get_stop_times(date)
    return (st["departure_time"].dropna().min(), st["arrival_time"].dropna().max())


def stop_times_to_geojson(
    feed: "Feed",
    trip_ids: Iterable[str | None] = None,
) -> dict:
    """
    Return a GeoJSON FeatureCollection of Point features
    representing all the trip-stop pairs in ``feed.stop_times``.
    The coordinates reference system is the default one for GeoJSON,
    namely WGS84.

    For every trip, drop duplicate stop IDs within that trip.
    In particular, a looping trip will lack its final stop.

    If an iterable of trip IDs is given, then subset to those trips.
    If some of the given trip IDs are not found in the feed, then raise a ValueError.
    """
    from .stops import get_stops

    if trip_ids is None or not list(trip_ids):
        trip_ids = feed.trips.trip_id

    D = set(trip_ids) - set(feed.trips.trip_id)
    if D:
        raise ValueError(f"Trip IDs {D} not found in feed.")

    st = feed.stop_times.loc[lambda x: x.trip_id.isin(trip_ids)]

    g = (
        get_stops(feed, as_gdf=True)
        .loc[lambda x: x["stop_id"].isin(st["stop_id"].unique())]
        .merge(st)
        .sort_values(["trip_id", "stop_sequence"])
        .drop_duplicates(subset=["trip_id", "stop_id"])
    )

    return hp.drop_feature_ids(json.loads(g.to_json()))



"""
Constants useful across modules.
"""

import pandas as pd

# Record some data from the GTFS reference at https://gtfs.org/reference/static.
columns = ["table", "table_required", "column", "column_required", "dtype"]
rows = [
    ["agency", True, "agency_id", False, "str"],
    ["agency", True, "agency_name", True, "str"],
    ["agency", True, "agency_url", True, "str"],
    ["agency", True, "agency_timezone", True, "str"],
    ["agency", True, "agency_lang", False, "str"],
    ["agency", True, "agency_phone", False, "str"],
    ["agency", True, "agency_fare_url", False, "str"],
    ["agency", True, "agency_email", False, "str"],
    ["attributions", False, "attribution_id", False, "str"],
    ["attributions", False, "agency_id", False, "str"],
    ["attributions", False, "route_id", False, "str"],
    ["attributions", False, "trip_id", False, "str"],
    ["attributions", False, "organization_name", True, "str"],
    ["attributions", False, "is_producer", False, "int"],
    ["attributions", False, "is_operator", False, "int"],
    ["attributions", False, "is_authority", False, "int"],
    ["attributions", False, "attribution_url", False, "str"],
    ["attributions", False, "attribution_email", False, "str"],
    ["attributions", False, "attribution_phone", False, "str"],
    ["calendar", False, "service_id", True, "str"],
    ["calendar", False, "monday", True, "int"],
    ["calendar", False, "tuesday", True, "int"],
    ["calendar", False, "wednesday", True, "int"],
    ["calendar", False, "thursday", True, "int"],
    ["calendar", False, "friday", True, "int"],
    ["calendar", False, "saturday", True, "int"],
    ["calendar", False, "sunday", True, "int"],
    ["calendar", False, "start_date", True, "str"],
    ["calendar", False, "end_date", True, "str"],
    ["calendar_dates", False, "service_id", True, "str"],
    ["calendar_dates", False, "date", True, "str"],
    ["calendar_dates", False, "exception_type", True, "int"],
    ["fare_attributes", False, "fare_id", True, "str"],
    ["fare_attributes", False, "price", True, "float"],
    ["fare_attributes", False, "currency_type", True, "str"],
    ["fare_attributes", False, "payment_method", True, "int"],
    ["fare_attributes", False, "transfers", True, "int"],
    ["fare_attributes", False, "transfer_duration", False, "int"],
    ["fare_rules", False, "fare_id", True, "str"],
    ["fare_rules", False, "route_id", False, "str"],
    ["fare_rules", False, "origin_id", False, "str"],
    ["fare_rules", False, "destination_id", False, "str"],
    ["fare_rules", False, "contains_id", False, "str"],
    ["feed_info", False, "feed_publisher_name", True, "str"],
    ["feed_info", False, "feed_publisher_url", True, "str"],
    ["feed_info", False, "feed_lang", True, "str"],
    ["feed_info", False, "feed_start_date", False, "str"],
    ["feed_info", False, "feed_end_date", False, "str"],
    ["feed_info", False, "feed_version", False, "str"],
    ["frequencies", False, "trip_id", True, "str"],
    ["frequencies", False, "start_time", True, "str"],
    ["frequencies", False, "end_time", True, "str"],
    ["frequencies", False, "headway_secs", True, "int"],
    ["frequencies", False, "exact_times", False, "int"],
    ["routes", True, "route_id", True, "str"],
    ["routes", True, "agency_id", False, "str"],
    ["routes", True, "route_short_name", True, "str"],
    ["routes", True, "route_long_name", True, "str"],
    ["routes", True, "route_desc", False, "str"],
    ["routes", True, "route_type", True, "int"],
    ["routes", True, "route_url", False, "str"],
    ["routes", True, "route_color", False, "str"],
    ["routes", True, "route_text_color", False, "str"],
    ["shapes", False, "shape_id", True, "str"],
    ["shapes", False, "shape_pt_lat", True, "float"],
    ["shapes", False, "shape_pt_lon", True, "float"],
    ["shapes", False, "shape_pt_sequence", True, "int"],
    ["shapes", False, "shape_dist_traveled", False, "float"],
    ["stops", True, "stop_id", True, "str"],
    ["stops", True, "stop_code", False, "str"],
    ["stops", True, "stop_name", True, "str"],
    ["stops", True, "stop_desc", False, "str"],
    ["stops", True, "stop_lat", True, "float"],
    ["stops", True, "stop_lon", True, "float"],
    ["stops", True, "zone_id", False, "str"],
    ["stops", True, "stop_url", False, "str"],
    ["stops", True, "location_type", False, "int"],
    ["stops", True, "parent_station", False, "str"],
    ["stops", True, "stop_timezone", False, "str"],
    ["stops", True, "wheelchair_boarding", False, "int"],
    ["stop_times", True, "trip_id", True, "str"],
    ["stop_times", True, "arrival_time", True, "str"],
    ["stop_times", True, "departure_time", True, "str"],
    ["stop_times", True, "stop_id", True, "str"],
    ["stop_times", True, "stop_sequence", True, "int"],
    ["stop_times", True, "stop_headsign", False, "str"],
    ["stop_times", True, "pickup_type", False, "int"],
    ["stop_times", True, "drop_off_type", False, "int"],
    ["stop_times", True, "shape_dist_traveled", False, "float"],
    ["stop_times", True, "timepoint", False, "int"],
    ["transfers", False, "from_stop_id", True, "str"],
    ["transfers", False, "to_stop_id", True, "str"],
    ["transfers", False, "transfer_type", True, "int"],
    ["transfers", False, "min_transfer_time", False, "int"],
    ["trips", True, "route_id", True, "str"],
    ["trips", True, "service_id", True, "str"],
    ["trips", True, "trip_id", True, "str"],
    ["trips", True, "trip_headsign", False, "str"],
    ["trips", True, "trip_short_name", False, "str"],
    ["trips", True, "direction_id", False, "int"],
    ["trips", True, "block_id", False, "str"],
    ["trips", True, "shape_id", False, "str"],
    ["trips", True, "wheelchair_accessible", False, "int"],
    ["trips", True, "bikes_allowed", False, "int"],
]
GTFS_REF = pd.DataFrame(rows, columns=columns)

#: Columns that must be formatted as integers when outputting GTFS
INT_COLS = GTFS_REF.loc[GTFS_REF["dtype"] == "int", "column"].values.tolist()

#: Columns that must be read as strings by Pandas
STR_COLS = GTFS_REF.loc[GTFS_REF["dtype"] == "str", "column"].values.tolist()

DTYPE = {col: str for col in STR_COLS}

#: Valid distance units
DIST_UNITS = ["ft", "mi", "m", "km"]

#: Primary feed attributes
FEED_ATTRS_1 = [
    "agency",
    "attributions",
    "calendar",
    "calendar_dates",
    "fare_attributes",
    "fare_rules",
    "feed_info",
    "frequencies",
    "routes",
    "shapes",
    "stops",
    "stop_times",
    "trips",
    "transfers",
    "dist_units",
]

#: Secondary feed attributes; derived from primary ones
FEED_ATTRS_2 = ["_trips_i", "_calendar_i", "_calendar_dates_i"]

#:
FEED_ATTRS = FEED_ATTRS_1 + FEED_ATTRS_2

#: WGS84 coordinate reference system for Geopandas
WGS84 = "EPSG:4326"

#: Colorbrewer 8-class Set2 colors
COLORS_SET2 = [
    "#66c2a5",
    "#fc8d62",
    "#8da0cb",
    "#e78ac3",
    "#a6d854",
    "#ffd92f",
    "#e5c494",
    "#b3b3b3",
]



import pandas as pd
import numpy as np
import pytest
import geopandas as gpd
import folium as fl

from .context import (
    gtfs_kit,
    DATA_DIR,
    cairns,
    cairns_shapeless,
    cairns_dates,
    cairns_trip_stats,
)
from gtfs_kit import constants as cs
from gtfs_kit import trips as gkt
from gtfs_kit import calendar as gkc
from gtfs_kit import helpers as gkh
from gtfs_kit import stop_times as gks


def test_is_active_trip():
    feed = cairns.copy()
    trip_id = "CNS2014-CNS_MUL-Weekday-00-4165878"
    date1 = "20140526"
    date2 = "20120322"
    assert gkt.is_active_trip(feed, trip_id, date1)
    assert not gkt.is_active_trip(feed, trip_id, date2)

    trip_id = "CNS2014-CNS_MUL-Sunday-00-4165971"
    date1 = "20140601"
    date2 = "20120602"
    assert gkt.is_active_trip(feed, trip_id, date1)
    assert not gkt.is_active_trip(feed, trip_id, date2)


def test_get_trips():
    feed = cairns.copy()
    date = cairns_dates[0]
    trips1 = gkt.get_trips(feed, date)
    # Should be a data frame
    assert isinstance(trips1, pd.core.frame.DataFrame)
    # Should have the correct shape
    assert trips1.shape[0] <= feed.trips.shape[0]
    assert trips1.shape[1] == feed.trips.shape[1]
    # Should have correct columns
    assert set(trips1.columns) == set(feed.trips.columns)

    trips2 = gkt.get_trips(feed, date, "07:30:00")
    # Should be a data frame
    assert isinstance(trips2, pd.core.frame.DataFrame)
    # Should have the correct shape
    assert trips2.shape[0] <= trips2.shape[0]
    assert trips2.shape[1] == trips1.shape[1]
    # Should have correct columns
    assert set(trips2.columns) == set(feed.trips.columns)

    feed = cairns.copy()
    g = gkt.get_trips(feed, as_gdf=True)
    assert g.crs == cs.WGS84
    assert g.shape[0] == feed.trips.shape[0]
    assert set(g.columns) == set(feed.trips.columns) | {"geometry"}

    with pytest.raises(ValueError):
        gkt.get_trips(cairns_shapeless, as_gdf=True)


def test_compute_trip_activity():
    feed = cairns.copy()
    dates = gkc.get_first_week(feed)
    trips_activity = gkt.compute_trip_activity(feed, dates)
    # Should be a data frame
    assert isinstance(trips_activity, pd.core.frame.DataFrame)
    # Should have the correct shape
    assert trips_activity.shape[0] == feed.trips.shape[0]
    assert trips_activity.shape[1] == 1 + len(dates)
    # Date columns should contain only zeros and ones
    assert set(trips_activity[dates].values.flatten()) == {0, 1}


def test_compute_busiest_date():
    feed = cairns.copy()
    dates = gkc.get_first_week(feed)[:1]
    date = gkt.compute_busiest_date(feed, dates + ["999"])
    # Busiest day should lie in first week
    assert date in dates


def test_compute_trip_stats():
    feed = cairns.copy()
    n = 3
    rids = feed.routes.route_id.loc[:n]
    trip_stats = gkt.compute_trip_stats(feed, route_ids=rids)

    # Should be a data frame with the correct number of rows
    trip_subset = feed.trips.loc[lambda x: x["route_id"].isin(rids)].copy()
    assert isinstance(trip_stats, pd.core.frame.DataFrame)
    assert trip_stats.shape[0] == trip_subset.shape[0]

    # Should contain the correct columns
    expect_cols = set(
        [
            "trip_id",
            "direction_id",
            "route_id",
            "route_short_name",
            "route_type",
            "shape_id",
            "num_stops",
            "start_time",
            "end_time",
            "start_stop_id",
            "end_stop_id",
            "distance",
            "duration",
            "speed",
            "is_loop",
        ]
    )
    assert set(trip_stats.columns) == expect_cols

    # Distance units should be correct
    d1 = trip_stats.distance.iat[0]  # km
    trip_stats_2 = gkt.compute_trip_stats(feed.convert_dist("ft"), route_ids=rids)
    d2 = trip_stats_2.distance.iat[0]  # mi
    f = gkh.get_convert_dist("km", "mi")
    assert np.allclose(f(d1), d2)

    # Shapeless feeds should have null entries for distance column
    feed = cairns_shapeless.copy()
    trip_stats = gkt.compute_trip_stats(feed, route_ids=rids)
    assert len(trip_stats["distance"].unique()) == 1
    assert np.isnan(trip_stats["distance"].unique()[0])

    # Should contain the correct trips
    get_trips = set(trip_stats["trip_id"].values)
    expect_trips = set(trip_subset["trip_id"].values)
    assert get_trips == expect_trips

    # Missing the optional ``direction_id`` column in ``feed.trips``
    # should give ``direction_id`` column in stats with all NaNs
    feed = cairns.copy()
    del feed.trips["direction_id"]
    trip_stats = gkt.compute_trip_stats(feed, route_ids=rids)
    assert set(trip_stats.columns) == expect_cols
    assert trip_stats.direction_id.isnull().all()

    # Missing the optional ``shape_id`` column in ``feed.trips``
    # should give ``shape_id`` column in stats with all NaNs
    feed = cairns.copy()
    del feed.trips["shape_id"]
    trip_stats = gkt.compute_trip_stats(feed, route_ids=rids)
    assert set(trip_stats.columns) == expect_cols
    assert trip_stats.shape_id.isnull().all()


@pytest.mark.slow
def test_locate_trips():
    feed = cairns.copy()
    feed = gks.append_dist_to_stop_times(feed)
    date = cairns_dates[0]
    times = ["08:00:00"]
    f = gkt.locate_trips(feed, date, times)
    g = gkt.get_trips(feed, date, times[0])
    # Should be a data frame
    assert isinstance(f, pd.core.frame.DataFrame)
    # Should have the correct number of rows
    assert f.shape[0] == g.shape[0]
    # Should have the correct columns
    expect_cols = set(
        [
            "route_id",
            "trip_id",
            "direction_id",
            "shape_id",
            "time",
            "rel_dist",
            "lon",
            "lat",
        ]
    )
    assert set(f.columns) == expect_cols

    # Missing feed.trips.shape_id should raise an error
    feed = cairns_shapeless.copy()
    del feed.trips["shape_id"]
    with pytest.raises(ValueError):
        gkt.locate_trips(feed, date, times)


def test_trips_to_geojson():
    feed = cairns.copy()
    trip_ids = feed.trips.trip_id.loc[:1]
    n = len(trip_ids)
    gj = gkt.trips_to_geojson(feed, trip_ids)
    assert len(gj["features"]) == n

    gj = gkt.trips_to_geojson(feed, trip_ids, include_stops=True)
    k = (
        feed.stop_times.loc[lambda x: x.trip_id.isin(trip_ids)]
        .drop_duplicates(subset=["trip_id", "stop_id"])
        .shape[0]
    )
    assert len(gj["features"]) == n + k

    with pytest.raises(ValueError):
        gkt.trips_to_geojson(cairns_shapeless)

    with pytest.raises(ValueError):
        gkt.trips_to_geojson(cairns, trip_ids=["bingo"])


def test_map_trips():
    feed = cairns.copy()
    tids = feed.trips["trip_id"].values[:2]
    m = gkt.map_trips(feed, tids, show_stops=True)
    assert isinstance(m, fl.Map)



from pandas.testing import assert_frame_equal, assert_series_equal
import numpy as np
import pytest

from .context import sample, gtfs_kit
from gtfs_kit import cleaners as gkc
from gtfs_kit import helpers as hp


def test_clean_column_names():
    f = sample.routes.copy()
    g = gkc.clean_column_names(f)
    assert_frame_equal(f, g)

    f = sample.routes.copy()
    f[" route_id  "] = f["route_id"].copy()
    del f["route_id"]
    g = gkc.clean_column_names(f)
    assert "route_id" in g.columns
    assert " route_id  " not in g.columns


def test_clean_ids():
    f1 = sample.copy()
    f1.routes.loc[0, "route_id"] = "  ho   ho ho "
    f2 = gkc.clean_ids(f1)
    expect_rid = "ho_ho_ho"
    assert f2.routes.loc[0, "route_id"] == expect_rid

    f3 = gkc.clean_ids(f2)
    assert f3 == f2


def test_extend_id():
    f1 = sample.copy()
    assert len(f1.trips.set_index("route_id").loc["AAMV"]) == 4

    f2 = gkc.extend_id(f1, "route_id", "prefix_")
    assert f2.routes.route_id.str.startswith("prefix_").all()
    assert len(f2.trips.set_index("route_id").loc["prefix_AAMV"]) == 4

    f3 = gkc.extend_id(f2, "route_id", "_suffix", prefix=False)
    assert f3.routes.route_id.str.endswith("_suffix").all()
    assert len(f3.trips.set_index("route_id").loc["prefix_AAMV_suffix"]) == 4

    with pytest.raises(ValueError):
        gkc.extend_id(f2, "direction_id", "_suffix", prefix=False)


def test_clean_times():
    f1 = sample.copy()
    f1.stop_times["departure_time"].iat[0] = "7:00:00"
    f1.frequencies["start_time"].iat[0] = "7:00:00"
    f2 = gkc.clean_times(f1)
    assert f2.stop_times["departure_time"].iat[0] == "07:00:00"
    assert f2.frequencies["start_time"].iat[0] == "07:00:00"


def test_clean_route_short_names():
    f1 = sample.copy()

    # Should have no effect on a fine feed
    f2 = gkc.clean_route_short_names(f1)
    assert_series_equal(f2.routes["route_short_name"], f1.routes["route_short_name"])

    # Make route short name duplicates
    f1.routes.loc[1:5, "route_short_name"] = np.nan
    f1.routes.loc[6:, "route_short_name"] = "  he llo  "
    f2 = gkc.clean_route_short_names(f1)
    # Should have unique route short names
    assert f2.routes["route_short_name"].nunique() == f2.routes.shape[0]
    # NaNs should be replaced by n/a and route IDs
    expect_rsns = ("n/a-" + sample.routes.iloc[1:5]["route_id"]).tolist()
    assert f2.routes.iloc[1:5]["route_short_name"].values.tolist() == expect_rsns
    # Should have names without leading or trailing whitespace
    assert not f2.routes["route_short_name"].str.startswith(" ").any()
    assert not f2.routes["route_short_name"].str.endswith(" ").any()


def test_drop_zombies():
    # Should have no effect on sample feed
    f1 = sample.copy()
    f2 = gkc.drop_zombies(f1)
    assert_frame_equal(f2.routes, f1.routes)

    # Should drop stops with no stop times
    f1 = sample.copy()
    f1.stops["location_type"] = np.nan
    stop_id = f1.stops.stop_id.iat[0]
    st = f1.stop_times.copy()
    st = st.loc[lambda x: x.stop_id != stop_id]
    f1.stop_times = st
    f2 = gkc.drop_zombies(f1)
    assert stop_id not in f2.stops.stop_id.values

    f2 = gkc.drop_zombies(f1)
    assert_frame_equal(f2.routes, f1.routes)

    # Create undefined parent stations
    f1 = sample.copy()
    f1.stops["parent_station"] = "bingo"
    f2 = gkc.drop_zombies(f1)
    assert f2.stops.parent_station.isna().all()

    # Create all zombie trips for one route
    rid = f1.routes["route_id"].iat[0]
    cond = f1.trips["route_id"] == rid
    f1.trips.loc[cond, "trip_id"] = "hoopla"
    f2 = gkc.drop_zombies(f1)
    # Trips should be gone
    assert "hoopla" not in f2.trips["trip_id"]
    # Route should be gone
    assert rid not in f2.routes["route_id"]


def test_build_aggregate_routes_dict():
    routes = sample.routes.copy()
    # Equalize all route short names
    routes["route_short_name"] = "bingo"
    nid_by_oid = gkc.build_aggregate_routes_dict(routes, route_id_prefix="bongo_")
    assert set(nid_by_oid.values()) == {"bongo_0"}


def test_aggregate_routes():
    feed1 = sample.copy()
    # Equalize all route short names
    feed1.routes["route_short_name"] = "bingo"
    feed2 = gkc.aggregate_routes(feed1)

    # feed2 should have only one route ID
    assert feed2.routes.shape[0] == 1

    # Feeds should have same trip DataFrames excluding route IDs
    feed1.trips["route_id"] = feed2.trips["route_id"]
    assert hp.almost_equal(feed1.trips, feed2.trips)

    # Feeds should have same fare rules DataFrames excluding route IDs
    feed1.fare_rules["route_id"] = feed2.fare_rules["route_id"]
    assert hp.almost_equal(feed1.fare_rules, feed2.fare_rules)

    # Feeds should have equal attributes excluding routes, trips, and fare rules
    # DataFrames
    feed2.routes = feed1.routes
    feed2.trips = feed1.trips
    feed2.fare_rules = feed1.fare_rules
    assert feed1 == feed2


def test_build_aggregate_stops_dict():
    stops = sample.stops.copy()
    # Equalize all stop codes
    stops["stop_code"] = "bingo"
    nid_by_oid = gkc.build_aggregate_stops_dict(stops, stop_id_prefix="bongo_")
    assert set(nid_by_oid.values()) == {"bongo_0"}


def test_aggregate_stops():
    feed1 = sample.copy()
    # Equalize all stop codes
    feed1.stops["stop_code"] = "bingo"
    feed2 = gkc.aggregate_stops(feed1)

    # feed2 should have only one stop ID
    assert feed2.stops.shape[0] == 1

    # Feeds should have same stop times, excluding stop IDs
    feed1.stop_times["stop_id"] = feed2.stop_times.stop_id
    assert hp.almost_equal(feed1.stop_times, feed2.stop_times)

    # Feeds should have equal attributes excluding
    # stops stop times DataFrames
    feed2.stops = feed1.stops
    feed2.stop_times = feed1.stop_times
    assert feed1 == feed2


def test_clean():
    f1 = sample.copy()
    rid = f1.routes["route_id"].iat[0]
    f1.routes["route_id"].iat[0] = " " + rid + "   "
    f2 = gkc.clean(f1)
    assert f2.routes["route_id"].iat[0] == rid
    assert_frame_equal(f2.trips, sample.trips)


def test_drop_invalid_columns():
    f1 = sample.copy()
    f1.routes["bingo"] = "bongo"
    f1.trips["wingo"] = "wongo"
    f2 = gkc.drop_invalid_columns(f1)
    assert f2 == sample



"""
Functions about miscellany.
"""

from __future__ import annotations
from typing import TYPE_CHECKING

import pandas as pd
import numpy as np
import shapely.geometry as sg
import geopandas as gpd

from . import helpers as hp
from . import constants as cs

# Help mypy but avoid circular imports
if TYPE_CHECKING:
    from .feed import Feed


def summarize(feed: "Feed", table: str | None = None) -> pd.DataFrame:
    """
    Return a DataFrame summarizing all GTFS tables in the given feed
    or in the given table if specified.

    The resulting DataFrame has the following columns.

    - ``'table'``: name of the GTFS table, e.g. ``'stops'``
    - ``'column'``: name of a column in the table,
      e.g. ``'stop_id'``
    - ``'num_values'``: number of values in the column
    - ``'num_nonnull_values'``: number of nonnull values in the
      column
    - ``'num_unique_values'``: number of unique values in the
      column, excluding null values
    - ``'min_value'``: minimum value in the column
    - ``'max_value'``: maximum value in the column

    If the table is not in the feed, then return an empty DataFrame
    If the table is not valid, raise a ValueError
    """
    gtfs_tables = cs.GTFS_REF.table.unique()

    if table is not None:
        if table not in gtfs_tables:
            raise ValueError(f"{table} is not a GTFS table")
        else:
            tables = [table]
    else:
        tables = gtfs_tables

    frames = []
    for table in tables:
        f = getattr(feed, table)
        if f is None:
            continue

        def my_agg(col):
            d = {}
            d["column"] = col.name
            d["num_values"] = col.size
            d["num_nonnull_values"] = col.count()
            d["num_unique_values"] = col.nunique()
            d["min_value"] = col.dropna().min()
            d["max_value"] = col.dropna().max()
            return pd.Series(d)

        g = f.apply(my_agg).T.reset_index(drop=True)
        g["table"] = table
        frames.append(g)

    cols = [
        "table",
        "column",
        "num_values",
        "num_nonnull_values",
        "num_unique_values",
        "min_value",
        "max_value",
    ]

    if not frames:
        f = pd.DataFrame()
    else:
        f = pd.concat(frames)
        # Rearrange columns
        f = f[cols].copy()

    return f


def describe(feed: "Feed", sample_date: str | None = None) -> pd.DataFrame:
    """
    Return a DataFrame of various feed indicators and values,
    e.g. number of routes.
    Specialize some those indicators to the given YYYYMMDD sample date string,
    e.g. number of routes active on the date.

    The resulting DataFrame has the columns

    - ``'indicator'``: string; name of an indicator, e.g. 'num_routes'
    - ``'value'``: value of the indicator, e.g. 27

    """
    from . import calendar as cl

    d = dict()
    dates = cl.get_dates(feed)
    d["agencies"] = feed.agency["agency_name"].tolist()
    d["timezone"] = feed.agency["agency_timezone"].iat[0]
    d["start_date"] = dates[0]
    d["end_date"] = dates[-1]
    d["num_routes"] = feed.routes.shape[0]
    d["num_trips"] = feed.trips.shape[0]
    d["num_stops"] = feed.stops.shape[0]
    if feed.shapes is not None:
        d["num_shapes"] = feed.shapes["shape_id"].nunique()
    else:
        d["num_shapes"] = 0

    if sample_date is None or sample_date not in feed.get_dates():
        sample_date = cl.get_first_week(feed)[3]
    d["sample_date"] = sample_date
    d["num_routes_active_on_sample_date"] = feed.get_routes(sample_date).shape[0]
    trips = feed.get_trips(sample_date)
    d["num_trips_active_on_sample_date"] = trips.shape[0]
    d["num_stops_active_on_sample_date"] = feed.get_stops(sample_date).shape[0]
    f = pd.DataFrame(list(d.items()), columns=["indicator", "value"])

    return f


def assess_quality(feed: "Feed") -> pd.DataFrame:
    """
    Return a DataFrame of various feed indicators and values,
    e.g. number of trips missing shapes.

    The resulting DataFrame has the columns

    - ``'indicator'``: string; name of an indicator, e.g. 'num_routes'
    - ``'value'``: value of the indicator, e.g. 27

    This function is odd but useful for seeing roughly how broken a feed is
    This function is not a GTFS validator.
    """
    d = dict()

    # Count duplicate route short names
    r = feed.routes
    dup = r.duplicated(subset=["route_short_name"])
    n = dup[dup].count()
    d["num_route_short_names_duplicated"] = n
    d["frac_route_short_names_duplicated"] = n / r.shape[0]

    # Count stop times missing shape_dist_traveled values
    st = feed.stop_times.sort_values(["trip_id", "stop_sequence"])
    if "shape_dist_traveled" in st.columns:
        # Count missing distances
        n = st[st["shape_dist_traveled"].isnull()].shape[0]
        d["num_stop_time_dists_missing"] = n
        d["frac_stop_time_dists_missing"] = n / st.shape[0]
    else:
        d["num_stop_time_dists_missing"] = st.shape[0]
        d["frac_stop_time_dists_missing"] = 1

    # Count direction_ids missing
    t = feed.trips
    if "direction_id" in t.columns:
        n = t[t["direction_id"].isnull()].shape[0]
        d["num_direction_ids_missing"] = n
        d["frac_direction_ids_missing"] = n / t.shape[0]
    else:
        d["num_direction_ids_missing"] = t.shape[0]
        d["frac_direction_ids_missing"] = 1

    # Count trips missing shapes
    if feed.shapes is not None:
        n = t[t["shape_id"].isnull()].shape[0]
    else:
        n = t.shape[0]
    d["num_trips_missing_shapes"] = n
    d["frac_trips_missing_shapes"] = n / t.shape[0]

    # Count missing departure times
    n = st[st["departure_time"].isnull()].shape[0]
    d["num_departure_times_missing"] = n
    d["frac_departure_times_missing"] = n / st.shape[0]

    # Count missing first departure times missing
    g = st.groupby("trip_id").first().reset_index()
    n = g[g["departure_time"].isnull()].shape[0]
    d["num_first_departure_times_missing"] = n
    d["frac_first_departure_times_missing"] = n / st.shape[0]

    # Count missing last departure times
    g = st.groupby("trip_id").last().reset_index()
    n = g[g["departure_time"].isnull()].shape[0]
    d["num_last_departure_times_missing"] = n
    d["frac_last_departure_times_missing"] = n / st.shape[0]

    # Opine
    if (
        (d["frac_first_departure_times_missing"] >= 0.1)
        or (d["frac_last_departure_times_missing"] >= 0.1)
        or d["frac_trips_missing_shapes"] >= 0.8
    ):
        d["assessment"] = "bad feed"
    elif (
        d["frac_direction_ids_missing"]
        or d["frac_stop_time_dists_missing"]
        or d["num_route_short_names_duplicated"]
    ):
        d["assessment"] = "probably a fixable feed"
    else:
        d["assessment"] = "good feed"

    f = pd.DataFrame(list(d.items()), columns=["indicator", "value"])

    return f


def convert_dist(feed: "Feed", new_dist_units: str) -> "Feed":
    """
    Convert the distances recorded in the ``shape_dist_traveled``
    columns of the given Feed to the given distance units.
    New distance units must lie in :const:`.constants.DIST_UNITS`.
    Return the resulting Feed.
    """
    feed = feed.copy()

    if feed.dist_units == new_dist_units:
        # Nothing to do
        return feed

    old_dist_units = feed.dist_units
    feed.dist_units = new_dist_units

    converter = hp.get_convert_dist(old_dist_units, new_dist_units)

    if hp.is_not_null(feed.stop_times, "shape_dist_traveled"):
        feed.stop_times["shape_dist_traveled"] = feed.stop_times[
            "shape_dist_traveled"
        ].map(converter)

    if hp.is_not_null(feed.shapes, "shape_dist_traveled"):
        feed.shapes["shape_dist_traveled"] = feed.shapes["shape_dist_traveled"].map(
            converter
        )

    return feed


def compute_feed_stats_0(
    feed: "Feed", trip_stats_subset: pd.DataFrame, *, split_route_types=False
) -> pd.DataFrame:
    """
    Helper function for :func:`compute_feed_stats`.
    """
    ts = trip_stats_subset.copy()
    stop_times = feed.stop_times.copy()

    # Convert timestrings to seconds for quicker calculations
    ts[["start_time", "end_time"]] = ts[["start_time", "end_time"]].map(
        hp.timestr_to_seconds
    )

    if split_route_types:
        # Compute stats
        stats_list = []
        for route_type, g in ts.groupby("route_type"):
            d = {}
            d["route_type"] = route_type
            d["num_stops"] = stop_times.loc[
                lambda x: x.trip_id.isin(g.trip_id), "stop_id"
            ].nunique()
            d["num_routes"] = g.route_id.nunique()
            d["num_trips"] = g.shape[0]
            d["num_trip_starts"] = g.start_time.count()
            d["num_trip_ends"] = g.loc[g.end_time < 24 * 3600, "end_time"].count()
            d["service_distance"] = g.distance.sum()
            d["service_duration"] = g.duration.sum()
            if d["service_distance"]:
                d["service_speed"] = d["service_distance"] / d["service_duration"]
            else:
                d["service_speed"] = 0

            # Compute peak stats, which is the slowest part
            active_trips = hp.get_active_trips_df(g[["start_time", "end_time"]])
            times, counts = (active_trips.index.values, active_trips.values)
            start, end = hp.get_peak_indices(times, counts)
            d["peak_num_trips"] = counts[start]
            d["peak_start_time"] = times[start]
            d["peak_end_time"] = times[end]

            stats_list.append(pd.Series(d))

        stats = pd.DataFrame(stats_list)

    else:
        # Compute stats
        d = {}
        d["num_stops"] = stop_times.stop_id.nunique()
        d["num_routes"] = ts.route_id.nunique()
        d["num_trips"] = ts.shape[0]
        d["num_trip_starts"] = ts.start_time.count()
        d["num_trip_ends"] = ts.loc[ts.end_time < 24 * 3600, "end_time"].count()
        d["service_distance"] = ts.distance.sum()
        d["service_duration"] = ts.duration.sum()
        d["service_speed"] = d["service_distance"] / d["service_duration"]

        # Compute peak stats, which is the slowest part
        active_trips = hp.get_active_trips_df(ts[["start_time", "end_time"]])
        times, counts = active_trips.index.values, active_trips.values
        start, end = hp.get_peak_indices(times, counts)
        d["peak_num_trips"] = counts[start]
        d["peak_start_time"] = times[start]
        d["peak_end_time"] = times[end]

        stats = pd.DataFrame(d, index=[0])

    # Convert seconds back to timestrings
    times = ["peak_start_time", "peak_end_time"]
    stats[times] = stats[times].map(lambda t: hp.timestr_to_seconds(t, inverse=True))

    return stats


def compute_feed_stats(
    feed: "Feed",
    trip_stats: pd.DataFrame,
    dates: list[str],
    *,
    split_route_types=False,
) -> pd.DataFrame:
    """
    Compute some stats for the given Feed, trip stats (in the format output by the
    function :func:`.trips.compute_trip_stats`) and dates (YYYYMMDD date stings).

    Return a DataFrame with the columns

    - ``'date'``
    - ``'route_type'`` (optional): presest if and only if ``split_route_types``
    - ``'num_stops'``: number of stops active on the date
    - ``'num_routes'``: number of routes active on the date
    - ``'num_trips'``: number of trips that start on the date
    - ``'num_trip_starts'``: number of trips with nonnull start
      times on the date
    - ``'num_trip_ends'``: number of trips with nonnull start times
      and nonnull end times on the date, ignoring trips that end
      after 23:59:59 on the date
    - ``'peak_num_trips'``: maximum number of simultaneous trips in
      service on the date
    - ``'peak_start_time'``: start time of first longest period
      during which the peak number of trips occurs on the date
    - ``'peak_end_time'``: end time of first longest period during
      which the peak number of trips occurs on the date
    - ``'service_distance'``: sum of the service distances for the
      active routes on the date;
      measured in kilometers if ``feed.dist_units`` is metric;
      otherwise measured in miles;
      contains all ``np.nan`` entries if ``feed.shapes is None``
    - ``'service_duration'``: sum of the service durations for the
      active routes on the date; measured in hours
    - ``'service_speed'``: service_distance/service_duration on the
      date

    Exclude dates with no active stops, which could yield the empty DataFrame.

    The route and trip stats for date d contain stats for trips that
    start on date d only and ignore trips that start on date d-1 and
    end on date d.
    """
    dates = feed.subset_dates(dates)
    if not dates:
        return pd.DataFrame()

    # Collect stats for each date,
    # memoizing stats the sequence of trip IDs active on the date
    # to avoid unnecessary recomputations.
    # Store in a dictionary of the form
    # trip ID sequence -> stats DataFarme.
    stats_by_ids = {}

    activity = feed.compute_trip_activity(dates)

    frames = []
    for date in dates:
        ids = tuple(activity.loc[activity[date] > 0, "trip_id"])
        if ids in stats_by_ids:
            stats = (
                stats_by_ids[ids]
                # Assign date
                .assign(date=date)
            )
        elif ids:
            # Compute stats
            ts = trip_stats.loc[lambda x: x.trip_id.isin(ids)].copy()
            stats = (
                compute_feed_stats_0(feed, ts, split_route_types=split_route_types)
                # Assign date
                .assign(date=date)
            )

            # Memoize stats
            stats_by_ids[ids] = stats
        else:
            stats = pd.DataFrame()

        frames.append(stats)

    # Assemble stats into a single DataFrame
    return pd.concat(frames)


def compute_feed_time_series(
    feed: "Feed",
    trip_stats: pd.DataFrame,
    dates: list[str],
    freq: str = "5Min",
    *,
    split_route_types: bool = False,
) -> pd.DataFrame:
    """
    Compute some feed stats in time series form for the given dates
    (YYYYMMDD date strings) and trip stats (of the form output by the function
    :func:`.trips.compute_trip_stats`).
    Use the given Pandas frequency string ``freq`` to specify the frequency of the
    resulting time series, e.g. '5Min'; highest frequency allowable is one minute
    ('1Min').
    If ``split_route_types``, then split stats by route type; otherwise don't

    Return a time series DataFrame with a datetime index across the given dates sampled
    at the given frequency across the given dates.
    The columns are

    - ``'num_trips'``: number of trips in service during during the
      time period
    - ``'num_trip_starts'``: number of trips with starting during the
      time period
    - ``'num_trip_ends'``: number of trips ending during the
      time period, ignoring the trips the end past midnight
    - ``'service_distance'``: distance traveled during the time
      period by all trips active during the time period;
      measured in kilometers if ``feed.dist_units`` is metric;
      otherwise measured in miles;
      contains all ``np.nan`` entries if ``feed.shapes is None``
    - ``'service_duration'``: duration traveled during the time
      period by all trips active during the time period;
      measured in hours
    - ``'service_speed'``: ``service_distance/service_duration``

    Exclude dates that lie outside of the Feed's date range.
    If all the dates given lie outside of the Feed's date range,
    then return an empty DataFrame with the specified columns.

    If ``split_route_types``, then multi-index the columns with

    - top level: name is ``'indicator'``; values are
      ``'num_trip_starts'``, ``'num_trip_ends'``, ``'num_trips'``,
      ``'service_distance'``, ``'service_duration'``, and
      ``'service_speed'``
    - bottom level: name is ``'route_type'``; values are route type values

    If all dates lie outside the Feed's date range, then return an
    empty DataFrame
    """
    rts = feed.compute_route_time_series(trip_stats, dates, freq=freq)
    if rts.empty:
        return pd.DataFrame()

    cols = [
        "num_trip_starts",
        "num_trip_ends",
        "num_trips",
        "service_distance",
        "service_duration",
    ]

    if split_route_types:
        f = (
            hp.unstack_time_series(rts)
            .merge(feed.routes.filter(["route_id", "route_type"]), how="left")
            .groupby(["datetime", "indicator", "route_type"])
            .agg({"value": lambda x: x.sum(min_count=1)})  # All-NaNs should sum to NaN
            .reset_index()
            .pipe(hp.restack_time_series)
        )
    else:
        f = (
            pd.concat(
                [rts[col].sum(axis="columns", min_count=1) for col in cols],
                axis=1,
                keys=cols,
            )
            .sort_index(axis="columns")
            .rename_axis(index="datetime")
        )
        f.columns.name = "indicator"

        # Set time series frequency
        f.index.freq = freq

    # Calculate service speed
    f["service_speed"] = (f.service_distance / f.service_duration).fillna(
        f.service_distance
    )

    return f


def create_shapes(feed: "Feed", *, all_trips: bool = False) -> "Feed":
    """
    Given a feed, create a shape for every trip that is missing a
    shape ID.
    Do this by connecting the stops on the trip with straight lines.
    Return the resulting feed which has updated shapes and trips
    tables.

    If ``all_trips``, then create new shapes for all trips by
    connecting stops, and remove the old shapes.
    """
    feed = feed.copy()

    if all_trips:
        trip_ids = feed.trips["trip_id"]
    else:
        trip_ids = feed.trips[feed.trips["shape_id"].isnull()]["trip_id"]

    # Get stop times for given trips
    f = feed.stop_times[feed.stop_times["trip_id"].isin(trip_ids)][
        ["trip_id", "stop_sequence", "stop_id"]
    ]
    f = f.sort_values(["trip_id", "stop_sequence"])

    if f.empty:
        # Nothing to do
        return feed

    # Create new shape IDs for given trips.
    # To do this, collect unique stop sequences,
    # sort them to impose a canonical order, and
    # assign shape IDs to them
    stop_seqs = sorted(
        set(tuple(group["stop_id"].values) for trip, group in f.groupby("trip_id"))
    )
    ids = hp.make_ids(len(stop_seqs), "shape_")
    shape_by_stop_seq = {seq: ids[i] for i, seq in enumerate(stop_seqs)}

    # Assign these new shape IDs to given trips
    shape_by_trip = {
        trip: shape_by_stop_seq[tuple(group["stop_id"].values)]
        for trip, group in f.groupby("trip_id")
    }
    trip_cond = feed.trips["trip_id"].isin(trip_ids)
    feed.trips.loc[trip_cond, "shape_id"] = feed.trips.loc[trip_cond, "trip_id"].map(
        lambda x: shape_by_trip[x]
    )

    # Build new shapes for given trips
    G = [
        [shape, i, stop]
        for stop_seq, shape in shape_by_stop_seq.items()
        for i, stop in enumerate(stop_seq)
    ]
    g = pd.DataFrame(G, columns=["shape_id", "shape_pt_sequence", "stop_id"])
    g = g.merge(feed.stops[["stop_id", "stop_lon", "stop_lat"]]).sort_values(
        ["shape_id", "shape_pt_sequence"]
    )
    g = g.drop(["stop_id"], axis=1)
    g = g.rename(columns={"stop_lon": "shape_pt_lon", "stop_lat": "shape_pt_lat"})

    if feed.shapes is not None and not all_trips:
        # Update feed shapes with new shapes
        feed.shapes = pd.concat([feed.shapes, g], sort=False)
    else:
        # Create all new shapes
        feed.shapes = g

    return feed


def compute_bounds(feed: "Feed", stop_ids: list[str] | None = None) -> np.array:
    """
    Return the bounding box (Numpy array [min longitude, min latitude, max longitude,
    max latitude]) of the given Feed's stops or of the subset of stops
    specified by the given stop IDs.
    """
    from .stops import get_stops

    g = get_stops(feed, as_gdf=True)
    if stop_ids is not None:
        g = g.loc[lambda x: x["stop_id"].isin(stop_ids)]

    return g.total_bounds


def compute_convex_hull(feed: "Feed", stop_ids: list[str] | None = None) -> sg.Polygon:
    """
    Return a convex hull (Shapely Polygon) representing the convex hull of the given
    Feed's stops or of the subset of stops specified by the given stop IDs.
    """
    from .stops import get_stops

    g = get_stops(feed, as_gdf=True)
    if stop_ids is not None:
        g = g.loc[lambda x: x["stop_id"].isin(stop_ids)]

    return g.union_all().convex_hull


def compute_centroid(feed: "Feed", stop_ids: list[str] | None = None) -> sg.Point:
    """
    Return the centroid (Shapely Point) of the convex hull the given Feed's stops
    or of the subset of stops specified by the given stop IDs.
    """
    from .stops import get_stops

    g = get_stops(feed, as_gdf=True)
    if stop_ids is not None:
        g = g.loc[lambda x: x["stop_id"].isin(stop_ids)]

    return g.union_all().convex_hull.centroid


def restrict_to_dates(feed: "Feed", dates: list[str]) -> "Feed":
    """
    Build a new feed by restricting this one to only the stops,
    trips, shapes, etc. active on at least one of the given dates
    (YYYYMMDD strings).
    Return the resulting feed, which will have empty non-agency tables
    if no trip is active on any of the given dates.
    """
    # Initialize the new feed as the old feed.
    # Restrict its DataFrames below.
    feed = feed.copy()

    # Get every trip that is active on at least one of the dates
    trip_activity = feed.compute_trip_activity(dates)
    if trip_activity.empty:
        trip_ids = []
    else:
        trip_ids = trip_activity.loc[
            lambda x: x.filter(dates).sum(axis=1) > 0,
            "trip_id",
        ]

    # Slice trips
    feed.trips = feed.trips.loc[lambda x: x.trip_id.isin(trip_ids)].copy()

    # Slice routes
    feed.routes = feed.routes.loc[lambda x: x.route_id.isin(feed.trips.route_id)].copy()

    # Slice stop times
    feed.stop_times = feed.stop_times.loc[lambda x: x.trip_id.isin(trip_ids)].copy()

    # Slice stops
    stop_ids = feed.stop_times.stop_id.unique()
    f = feed.stops.copy()
    cond = f.stop_id.isin(stop_ids)
    if "location_type" in f.columns:
        cond |= ~f.location_type.isin([0, np.nan])
    feed.stops = f[cond].copy()

    # Slice calendar
    service_ids = feed.trips.service_id
    if feed.calendar is not None:
        feed.calendar = feed.calendar.loc[
            lambda x: x.service_id.isin(service_ids)
        ].copy()

    # Get agency for trips
    if "agency_id" in feed.routes.columns:
        agency_ids = feed.routes.agency_id
        if len(agency_ids):
            feed.agency = feed.agency.loc[lambda x: x.agency_id.isin(agency_ids)].copy()

    # Now for the optional files.
    # Get calendar dates for trips.
    if feed.calendar_dates is not None:
        feed.calendar_dates = feed.calendar_dates.loc[
            lambda x: x.service_id.isin(service_ids) & x.date.isin(dates)
        ].copy()

    # Get frequencies for trips
    if feed.frequencies is not None:
        feed.frequencies = feed.frequencies.loc[
            lambda x: x.trip_id.isin(trip_ids)
        ].copy()

    # Get shapes for trips
    if feed.shapes is not None:
        shape_ids = feed.trips.shape_id
        feed.shapes = feed.shapes.loc[lambda x: x.shape_id.isin(shape_ids)].copy()

    # Get transfers for stops
    if feed.transfers is not None:
        feed.transfers = feed.transfers.loc[
            lambda x: x.from_stop_id.isin(stop_ids) & x.to_stop_id.isin(stop_ids)
        ].copy()

    return feed


def restrict_to_routes(feed: "Feed", route_ids: list[str]) -> "Feed":
    """
    Build a new feed by restricting this one to only the stops,
    trips, shapes, etc. used by the routes with the given list of
    route IDs.
    Return the resulting feed.
    """
    # Initialize the new feed as the old feed.
    # Restrict its DataFrames below.
    feed = feed.copy()

    # Slice routes
    feed.routes = feed.routes.loc[lambda x: x.route_id.isin(route_ids)].copy()

    # Slice trips
    feed.trips = feed.trips.loc[lambda x: x.route_id.isin(route_ids)].copy()

    # Slice stop times
    trip_ids = feed.trips.trip_id
    feed.stop_times = feed.stop_times.loc[lambda x: x.trip_id.isin(trip_ids)].copy()

    # Slice stops
    stop_ids = feed.stop_times.stop_id.unique()
    f = feed.stops.copy()
    cond = f.stop_id.isin(stop_ids)
    if "location_type" in f.columns:
        cond |= ~f.location_type.isin([0, np.nan])
    feed.stops = f[cond].copy()

    # Slice calendar
    service_ids = feed.trips.service_id
    if feed.calendar is not None:
        feed.calendar = feed.calendar.loc[
            lambda x: x.service_id.isin(service_ids)
        ].copy()

    # Get agency for trips
    if "agency_id" in feed.routes.columns:
        agency_ids = feed.routes.agency_id
        if len(agency_ids):
            feed.agency = feed.agency.loc[lambda x: x.agency_id.isin(agency_ids)].copy()

    # Now for the optional files.
    # Get calendar dates for trips.
    if feed.calendar_dates is not None:
        feed.calendar_dates = feed.calendar_dates.loc[
            lambda x: x.service_id.isin(service_ids)
        ].copy()

    # Get frequencies for trips
    if feed.frequencies is not None:
        feed.frequencies = feed.frequencies.loc[
            lambda x: x.trip_id.isin(trip_ids)
        ].copy()

    # Get shapes for trips
    if feed.shapes is not None:
        shape_ids = feed.trips.shape_id
        feed.shapes = feed.shapes[lambda x: x.shape_id.isin(shape_ids)].copy()

    # Get transfers for stops
    if feed.transfers is not None:
        feed.transfers = feed.transfers.loc[
            lambda x: (x.from_stop_id.isin(stop_ids) & x.to_stop_id.isin(stop_ids))
        ].copy()

    return feed


def restrict_to_area(feed: "Feed", area: gpd.GeoDataFrame) -> "Feed":
    """
    Build a new feed by restricting this one to only the trips
    that have at least one stop intersecting the given GeoDataFrame of polygons,
    then restricting stops, routes, stop times, etc. to those
    associated with that subset of trips.
    Return the resulting feed.
    """
    from .stops import get_stops_in_area

    feed = feed.copy()

    # Get IDs of stops within the polygon
    stop_ids = get_stops_in_area(feed, area).stop_id

    # Get all trips with at least one of those stops
    st = feed.stop_times.copy()
    trip_ids = st.loc[lambda x: x.stop_id.isin(stop_ids), "trip_id"]
    feed.trips = feed.trips.loc[lambda x: x.trip_id.isin(trip_ids)].copy()

    # Get stop times for trips
    feed.stop_times = st.loc[lambda x: x.trip_id.isin(trip_ids)].copy()

    # Slice stops
    stop_ids = feed.stop_times.stop_id.unique()
    f = feed.stops.copy()
    cond = f.stop_id.isin(stop_ids)
    if "location_type" in f.columns:
        cond |= ~f.location_type.isin([0, np.nan])
    feed.stops = f[cond].copy()

    # Get routes for trips
    route_ids = feed.trips.route_id
    feed.routes = feed.routes.loc[lambda x: x.route_id.isin(route_ids)].copy()

    # Get calendar for trips
    service_ids = feed.trips.service_id
    if feed.calendar is not None:
        feed.calendar = feed.calendar.loc[
            lambda x: x.service_id.isin(service_ids)
        ].copy()

    # Get agency for trips
    if "agency_id" in feed.routes.columns:
        agency_ids = feed.routes.agency_id
        if len(agency_ids):
            feed.agency = feed.agency.loc[lambda x: x.agency_id.isin(agency_ids)].copy()

    # Now for the optional files.
    # Get calendar dates for trips.
    cd = feed.calendar_dates
    if cd is not None:
        feed.calendar_dates = cd.loc[lambda x: x.service_id.isin(service_ids)].copy()

    # Get frequencies for trips
    if feed.frequencies is not None:
        feed.frequencies = feed.frequencies.loc[
            lambda x: x.trip_id.isin(trip_ids)
        ].copy()

    # Get shapes for trips
    if feed.shapes is not None:
        shape_ids = feed.trips.shape_id
        feed.shapes = feed.shapes.loc[lambda x: x.shape_id.isin(shape_ids)].copy()

    # Get transfers for stops
    if feed.transfers is not None:
        t = feed.transfers
        feed.transfers = t.loc[
            lambda x: x.from_stop_id.isin(stop_ids) & x.to_stop_id.isin(stop_ids)
        ].copy()

    return feed


def compute_screen_line_counts(
    feed: "Feed", screen_lines: gpd.GeoDataFrame, dates: list[str]
) -> pd.DataFrame:
    """
    Find all the Feed trips active on the given YYYYMMDD dates whose shapes
    intersect the given GeoDataFrame of screen lines, that is, of straight WGS84
    LineStrings.
    Compute the intersection times and directions for each trip.

    Return a DataFrame with the columns

    - ``'date'``
    - ``'trip_id'``
    - ``'route_id'``
    - ``'route_short_name'``
    - ``'shape_id'``: shape ID of the trip
    - ``'screen_line_id'``: ID of the screen line as specified in ``screen_lines`` or as
      assigned after the fact.
    - ``'crossing_distance'``: distance (in the feed's distance units) along the trip
      shape of the screen line intersection
      ``'crossing_time'``: time that the trip's vehicle crosses
      the scren line; one trip could cross multiple times
    - ``'crossing_direction'``: 1 or -1; 1 indicates trip travel from the
      left side to the right side of the screen line;
      -1 indicates trip travel in the  opposite direction

    Notes:

    - Assume the Feed's stop times DataFrame has an accurate ``shape_dist_traveled``
      column.
    - Assume that trips travel in the same direction as their shapes, an assumption
      that is part of the GTFS.
    - Assume that the screen line is straight and simple.
    - Probably does not give correct results for trips with self-intersecting shapes.
    - The algorithm works as follows

        1. Find the trip shapes that intersect the screen lines.
        2. For each such shape and screen line, compute the intersection points,
           the distance of the point along the shape, and the orientation of the screen
           line relative to the shape.
        3. For each given date, restrict to trips active on the
           date and interpolate a stop time for the intersection point using
           the ``shape_dist_traveled`` column.
        4. Use that interpolated time as the crossing time of the trip vehicle.

    """
    from .shapes import get_shapes

    dates = feed.subset_dates(dates)
    if not dates:
        return pd.DataFrame()

    # Get shapes as GeoDataFrame
    shapes_g = get_shapes(feed, as_gdf=True, use_utm=True)

    # Convert screen lines to UTM
    crs = shapes_g.crs
    screen_lines = screen_lines.to_crs(crs)

    # Create screen line IDs if necessary
    n = screen_lines.shape[0]
    if "screen_line_id" not in screen_lines.columns:
        screen_lines["screen_line_id"] = hp.make_ids(n, "sl")

    # Make a vector in the direction of each screen line to calculate crossing
    # orientation. Does not work in case of a bent screen line.
    p1 = screen_lines["geometry"].map(lambda x: np.array(x.coords[0]))
    p2 = screen_lines["geometry"].map(lambda x: np.array(x.coords[-1]))
    screen_lines["screen_line_vector"] = p2 - p1

    # Get intersection points of shapes and screen lines
    g0 = (
        # Only keep shapes that intersect screen lines to reduce computations
        gpd.sjoin(shapes_g, screen_lines.filter(["screen_line_id", "geometry"])).merge(
            screen_lines, on="screen_line_id"
        )
        # Compute intersection points
        .assign(
            int_point=lambda x: gpd.GeoSeries(x["geometry_x"], crs=crs).intersection(
                gpd.GeoSeries(x["geometry_y"], crs=crs)
            )
        )
    )

    # Unpack multipoint intersections to yield a new GeoDataFrame
    records = []
    for row in g0.itertuples(index=False):
        if isinstance(row.int_point, sg.Point):
            intersections = [row.int_point]
        else:
            intersections = row.int_point.geoms
        for int_point in intersections:
            record = {
                "shape_id": row.shape_id,
                "screen_line_id": row.screen_line_id,
                "geometry": row.geometry_x,
                "int_point": int_point,
                "screen_line_vector": row.screen_line_vector,
            }
            records.append(record)

    g = gpd.GeoDataFrame.from_records(records).set_geometry("geometry")
    g.crs = crs

    # Get distance (in meters) of each intersection point along shape
    g["crossing_dist"] = g.apply(lambda x: x["geometry"].project(x.int_point), axis=1)

    # Build a tiny vector along each shape
    p2 = g.apply(
        lambda x: x["geometry"].interpolate(x["crossing_dist"] + 1), axis=1
    ).map(lambda x: np.array(x.coords[0]))
    p1 = g.int_point.map(lambda x: np.array(x.coords[0]))
    g["shape_vector"] = p2 - p1

    # Compute crossing direction by taking the vector cross product of
    # the shape vector and the screen line vector
    det = g.apply(
        lambda x: np.linalg.det(np.array([x.shape_vector, x.screen_line_vector])),
        axis=1,
    )
    g["crossing_direction"] = det.map(lambda x: 1 if x >= 0 else -1)

    # Convert to feed distance units
    converter = hp.get_convert_dist("m", feed.dist_units)
    g["crossing_dist"] = g["crossing_dist"].map(converter)

    # Summarize work so far into a lookup table
    h = (
        g.filter(["shape_id", "screen_line_id", "crossing_direction", "crossing_dist"])
        .set_index("shape_id")
        .sort_values(
            ["shape_id", "crossing_dist"]
        )  # Need this sorting for interpolation to work
    )

    # Get stop times of trips whose shapes lie in h
    st = (
        feed.trips.loc[lambda x: x["shape_id"].isin(h.index)]
        # Merge in route short names and stop times
        .merge(feed.routes[["route_id", "route_short_name"]]).merge(feed.stop_times)
        # Keep only non-NaN departure times
        .loc[lambda x: x["departure_time"].notna()]
        # Convert to seconds past midnight
        .assign(departure_time=lambda x: x["departure_time"].map(hp.timestr_to_seconds))
    )

    # Compute crossing times by date
    records = []
    ta = feed.compute_trip_activity(dates)
    for date in dates:
        # Subset to trips active on date and merge with g
        ids = ta.loc[lambda x: x[date] == 1, "trip_id"]
        f = st.loc[lambda x: x["trip_id"].isin(ids)].sort_values(
            ["trip_id", "shape_dist_traveled"]
        )  # Need this sorting for interpolation to work

        # Get crossing time for each trip
        for tid, group in f.groupby("trip_id"):
            sid = group["shape_id"].iat[0]
            dists = group["shape_dist_traveled"].values
            times = group["departure_time"].values
            crossing_dists = h.loc[[sid], "crossing_dist"].values
            crossing_times = np.interp(crossing_dists, dists, times)
            for i, row in enumerate(h.loc[[sid]].itertuples(index=False)):
                record = {
                    "date": date,
                    "trip_id": tid,
                    "route_id": group["route_id"].iat[0],
                    "route_short_name": group["route_short_name"].iat[0],
                    "shape_id": group["shape_id"].iat[0],
                    "screen_line_id": row.screen_line_id,
                    "crossing_direction": row.crossing_direction,
                    "crossing_distance": row.crossing_dist,
                    "crossing_time": crossing_times[i],
                }
                records.append(record)

    result = pd.DataFrame.from_records(records).assign(
        crossing_time=lambda x: x["crossing_time"].map(
            lambda x: hp.timestr_to_seconds(x, inverse=True)
        )
    )
    return result



import pytest
import pandas as pd
import numpy as np

from .context import gtfs_kit, DATA_DIR, sample, cairns, cairns_dates, cairns_trip_stats
from gtfs_kit import stop_times as gks


def test_get_stop_times():
    feed = cairns.copy()
    date = cairns_dates[0]
    f = gks.get_stop_times(feed, date)
    # Should be a data frame
    assert isinstance(f, pd.core.frame.DataFrame)
    # Should have a reasonable shape
    assert f.shape[0] <= feed.stop_times.shape[0]
    # Should have correct columns
    assert set(f.columns) == set(feed.stop_times.columns)


def test_get_start_and_end_times():
    feed = cairns.copy()
    date = cairns_dates[0]
    st = gks.get_stop_times(feed, date)
    times = gks.get_start_and_end_times(feed, date)
    # Should be strings
    for t in times:
        assert isinstance(t, str)
        # Should lie in stop times
        assert t in st[["departure_time", "arrival_time"]].values.flatten()

    # Should get null times in some cases
    times = gks.get_start_and_end_times(feed, "19690711")
    for t in times:
        assert pd.isnull(t)
    feed.stop_times["departure_time"] = np.nan
    times = gks.get_start_and_end_times(feed)
    assert pd.isnull(times[0])


@pytest.mark.slow
def test_append_dist_to_stop_times():
    feed1 = cairns.copy()
    st1 = feed1.stop_times
    feed2 = gks.append_dist_to_stop_times(feed1)
    st2 = feed2.stop_times

    # Check that colums of st2 equal the columns of st1 plus
    # a shape_dist_traveled column
    cols1 = st1.columns.values.tolist() + ["shape_dist_traveled"]
    cols2 = st2.columns.values.tolist()
    assert set(cols1) == set(cols2)

    # Check that within each trip the shape_dist_traveled column
    # is monotonically increasing
    for trip, group in st2.groupby("trip_id"):
        group = group.sort_values("stop_sequence")
        sdt = group.shape_dist_traveled.values.tolist()
        assert sdt == sorted(sdt)


def test_stop_times_to_geojson():
    feed = cairns.copy()
    trip_ids = feed.trips.trip_id.unique()[:2]
    collection = gks.stop_times_to_geojson(feed, trip_ids)
    assert isinstance(collection, dict)

    n = (
        feed.stop_times.loc[lambda x: x.trip_id.isin(trip_ids)]
        .drop_duplicates(subset=["trip_id", "stop_id"])
        .shape[0]
    )
    assert len(collection["features"]) == n

    with pytest.raises(ValueError):
        gks.stop_times_to_geojson(feed, ["bingo"])



import numpy as np
import pandas as pd

from .context import gtfs_kit, sample
from gtfs_kit import validators as gkv


def test_valid_str():
    assert gkv.valid_str("hello3")
    assert not gkv.valid_str(np.nan)
    assert not gkv.valid_str(" ")


def test_valid_time():
    assert gkv.valid_time("2:43:00")
    assert gkv.valid_time("22:43:00")
    assert gkv.valid_time("42:43:00")
    assert not gkv.valid_time("142:43:00")


def test_valid_date():
    assert gkv.valid_date("20140310")
    assert not gkv.valid_date("2014031")


def test_valid_timezone():
    assert gkv.valid_timezone("Africa/Abidjan")
    assert not gkv.valid_timezone("zoom")


def test_valid_lang():
    assert gkv.valid_lang("aa")
    assert not gkv.valid_lang("aaa")


def test_valid_currency():
    assert gkv.valid_currency("AED")
    assert not gkv.valid_currency("aed")


def test_valid_url():
    assert gkv.valid_url("http://www.example.com")
    assert not gkv.valid_url("www.example.com")


def test_valid_email():
    assert gkv.valid_email("a@b.c.com")
    assert not gkv.valid_email("a@b@c.com")


def test_valid_color():
    assert gkv.valid_color("00FFFF")
    assert not gkv.valid_color("0FF")
    assert not gkv.valid_color("GGFFFF")


def test_check_for_required_columns():
    assert not gkv.check_for_required_columns([], "routes", sample.routes)

    feed = sample.copy()
    del feed.routes["route_type"]
    assert gkv.check_for_required_columns([], "routes", feed.routes)


def test_check_for_invalid_columns():
    assert not gkv.check_for_invalid_columns([], "routes", sample.routes)

    feed = sample.copy()
    feed.routes["bingo"] = "snoop"
    assert gkv.check_for_invalid_columns([], "routes", feed.routes)


def test_check_table():
    feed = sample.copy()
    cond = feed.routes["route_id"].isnull()
    assert not gkv.check_table([], "routes", feed.routes, cond, "Bingo")
    assert gkv.check_table([], "routes", feed.routes, ~cond, "Bongo")


def test_check_column():
    feed = sample.copy()
    assert not gkv.check_column([], "agency", feed.agency, "agency_url", gkv.valid_url)
    feed.agency["agency_url"].iat[0] = "example.com"
    assert gkv.check_column([], "agency", feed.agency, "agency_url", gkv.valid_url)


def test_check_column_id():
    feed = sample.copy()
    assert not gkv.check_column_id([], "routes", feed.routes, "route_id")
    feed.routes["route_id"].iat[0] = np.nan
    assert gkv.check_column_id([], "routes", feed.routes, "route_id")


def test_check_column_linked_id():
    feed = sample.copy()
    assert not gkv.check_column_linked_id(
        [], "trips", feed.trips, "route_id", feed.routes
    )
    feed.trips["route_id"].iat[0] = "Hummus!"
    assert gkv.check_column_linked_id([], "trips", feed.trips, "route_id", feed.routes)


def test_format_problems():
    problems = [("ba", "da", "boom", "boom")]
    assert problems == gkv.format_problems(problems, as_df=False)

    e = gkv.format_problems(problems, as_df=True)
    assert isinstance(e, pd.DataFrame)
    assert e.columns.tolist() == ["type", "message", "table", "rows"]


def test_check_agency():
    assert not gkv.check_agency(sample)

    feed = sample.copy()
    feed.agency = None
    assert gkv.check_agency(feed)

    feed = sample.copy()
    del feed.agency["agency_name"]
    assert gkv.check_agency(feed)

    feed = sample.copy()
    feed.agency["b"] = 3
    assert gkv.check_agency(feed, include_warnings=True)

    feed = sample.copy()
    feed.agency = pd.concat([feed.agency, feed.agency.iloc[0]])
    assert gkv.check_agency(feed)

    feed = sample.copy()
    feed.agency["agency_name"] = ""
    assert gkv.check_agency(feed)

    for col in [
        "agency_timezone",
        "agency_url",
        "agency_fare_url",
        "agency_lang",
        "agency_phone",
        "agency_email",
    ]:
        feed = sample.copy()
        feed.agency[col] = ""
        assert gkv.check_agency(feed)


def test_check_attributions():
    assert not gkv.check_attributions(sample)

    feed = sample.copy()
    del feed.attributions["organization_name"]
    assert gkv.check_attributions(feed)

    feed = sample.copy()
    feed.attributions["route_id"] = "no such route ID"
    # Can't have invalid route_id
    assert gkv.check_attributions(feed)

    feed = sample.copy()
    feed.attributions["trip_id"] = "AB1"
    # Can't have both route_id and trip_id present
    assert gkv.check_attributions(feed)

    feed = sample.copy()
    feed.attributions["is_producer"] = 0
    feed.attributions["is_operator"] = 0
    feed.attributions["is_authority"] = 0
    # Can't have all these be zero
    assert gkv.check_attributions(feed)

    for col in [
        "attribution_url",
        "attribution_email",
    ]:
        feed = sample.copy()
        feed.attributions[col] = ""
        assert gkv.check_attributions(feed)


def test_check_calendar():
    assert not gkv.check_calendar(sample)
    assert gkv.check_calendar(sample, include_warnings=True)  # feed has expired

    feed = sample.copy()
    feed.calendar = None
    assert not gkv.check_calendar(feed)

    feed = sample.copy()
    del feed.calendar["service_id"]
    assert gkv.check_calendar(feed)

    feed = sample.copy()
    feed.calendar["yo"] = 3
    assert not gkv.check_calendar(feed)
    assert gkv.check_calendar(feed, include_warnings=True)

    feed = sample.copy()
    feed.calendar["service_id"].iat[0] = feed.calendar["service_id"].iat[1]
    assert gkv.check_calendar(feed)

    for col in [
        "monday",
        "tuesday",
        "wednesday",
        "thursday",
        "friday",
        "saturday",
        "sunday",
    ]:
        feed = sample.copy()
        feed.calendar[col].iat[0] = -1
        assert gkv.check_calendar(feed)

    for col in [
        "start_date",
        "end_date",
    ]:
        feed = sample.copy()
        feed.calendar[col].iat[0] = "bingo"
        assert gkv.check_calendar(feed)


def test_check_calendar_dates():
    assert not gkv.check_calendar_dates(sample)

    feed = sample.copy()
    feed.calendar_dates = None
    assert not gkv.check_calendar_dates(feed)

    feed = sample.copy()
    del feed.calendar_dates["service_id"]
    assert gkv.check_calendar_dates(feed)

    feed = sample.copy()
    feed.calendar_dates["yo"] = 3
    assert not gkv.check_calendar_dates(feed)
    assert gkv.check_calendar_dates(feed, include_warnings=True)

    feed = sample.copy()
    feed.calendar_dates["date"].iat[0] = "5"
    assert gkv.check_calendar_dates(feed)

    feed = sample.copy()
    feed.calendar_dates["exception_type"].iat[0] = -13
    assert gkv.check_calendar_dates(feed)


def test_check_fare_attributes():
    assert not gkv.check_fare_attributes(sample)

    feed = sample.copy()
    feed.fare_attributes = None
    assert not gkv.check_fare_attributes(feed)

    feed = sample.copy()
    del feed.fare_attributes["fare_id"]
    assert gkv.check_fare_attributes(feed)

    feed = sample.copy()
    feed.fare_attributes["yo"] = 3
    assert not gkv.check_fare_attributes(feed)
    assert gkv.check_fare_attributes(feed, include_warnings=True)

    feed = sample.copy()
    feed.fare_attributes["currency_type"] = "jubjub"
    assert gkv.check_fare_attributes(feed)

    for col in ["payment_method", "transfers", "transfer_duration"]:
        feed = sample.copy()
        feed.fare_attributes[col] = -7
        assert gkv.check_fare_attributes(feed)


def test_check_fare_rules():
    assert not gkv.check_fare_rules(sample)

    feed = sample.copy()
    feed.fare_rules = None
    assert not gkv.check_fare_rules(feed)

    feed = sample.copy()
    del feed.fare_rules["fare_id"]
    assert gkv.check_fare_rules(feed)

    feed = sample.copy()
    feed.fare_rules["yo"] = 3
    assert not gkv.check_fare_rules(feed)
    assert gkv.check_fare_rules(feed, include_warnings=True)

    for col in [
        "fare_id",
        "route_id",
        "origin_id",
        "destination_id",
        "contains_id",
    ]:
        feed = sample.copy()
        feed.fare_rules[col] = "tuberosity"
        print(col)
        print(feed.fare_rules)
        assert gkv.check_fare_rules(feed)


def test_check_feed_info():
    # Create feed_info table
    feed = sample.copy()
    columns = [
        "feed_publisher_name",
        "feed_publisher_url",
        "feed_lang",
        "feed_start_date",
        "feed_end_date",
        "feed_version",
    ]
    rows = [["slurp", "http://slurp.burp", "aa", "21110101", "21110102", "69"]]
    feed.feed_info = pd.DataFrame(rows, columns=columns)
    assert not gkv.check_feed_info(feed)

    feed1 = feed.copy()
    feed1.feed_info = None
    assert not gkv.check_feed_info(feed1)

    feed1 = feed.copy()
    del feed1.feed_info["feed_lang"]
    assert gkv.check_feed_info(feed1)

    feed1 = feed.copy()
    feed1.feed_info["yo"] = 3
    assert not gkv.check_feed_info(feed1)
    assert gkv.check_feed_info(feed1, include_warnings=True)

    for col in columns:
        feed1 = feed.copy()
        feed1.feed_info[col] = ""
        assert gkv.check_feed_info(feed1)


def test_check_frequencies():
    assert not gkv.check_frequencies(sample)

    feed = sample.copy()
    feed.frequencies = None
    assert not gkv.check_frequencies(feed)

    feed = sample.copy()
    del feed.frequencies["trip_id"]
    assert gkv.check_frequencies(feed)

    feed = sample.copy()
    feed.frequencies["yo"] = 3
    assert not gkv.check_frequencies(feed)
    assert gkv.check_frequencies(feed, include_warnings=True)

    feed = sample.copy()
    feed.frequencies["trip_id"].iat[0] = "ratatat"
    assert gkv.check_frequencies(feed)

    for col in ["start_time", "end_time"]:
        feed = sample.copy()
        feed.frequencies[col] = "07:00:00"
        assert gkv.check_frequencies(feed)

    for col in ["headway_secs", "exact_times"]:
        feed = sample.copy()
        feed.frequencies[col] = -7
        assert gkv.check_frequencies(feed)


def test_check_routes():
    assert not gkv.check_routes(sample)

    feed = sample.copy()
    feed.routes = None
    assert gkv.check_routes(feed)

    feed = sample.copy()
    del feed.routes["route_id"]
    assert gkv.check_routes(feed)

    feed = sample.copy()
    feed.routes["bingo"] = 3
    assert gkv.check_routes(feed, include_warnings=True)

    feed = sample.copy()
    feed.routes["route_id"].iat[0] = feed.routes["route_id"].iat[1]
    assert gkv.check_routes(feed)

    feed = sample.copy()
    feed.routes["agency_id"] = "Hubba hubba"
    assert gkv.check_routes(feed)

    feed = sample.copy()
    feed.routes["route_short_name"].iat[0] = ""
    assert gkv.check_routes(feed)

    feed = sample.copy()
    feed.routes["route_short_name"].iat[0] = ""
    feed.routes["route_long_name"].iat[0] = ""
    assert gkv.check_routes(feed)

    feed = sample.copy()
    feed.routes["route_type"].iat[0] = 8
    assert gkv.check_routes(feed)

    feed = sample.copy()
    feed.routes["route_color"].iat[0] = "FFF"
    assert gkv.check_routes(feed)

    feed = sample.copy()
    feed.routes["route_text_color"].iat[0] = "FFF"
    assert gkv.check_routes(feed)

    feed = sample.copy()
    feed.routes["route_short_name"].iat[1] = feed.routes["route_short_name"].iat[0]
    feed.routes["route_long_name"].iat[1] = feed.routes["route_long_name"].iat[0]
    assert not gkv.check_routes(feed)
    assert gkv.check_routes(feed, include_warnings=True)

    feed = sample.copy()
    feed.routes["route_id"].iat[0] = "Shwing"
    assert not gkv.check_routes(feed)
    assert gkv.check_routes(feed, include_warnings=True)

    feed = sample.copy()
    feed.agency = None
    assert gkv.check_routes(feed)


def test_check_shapes():
    assert not gkv.check_shapes(sample)

    # Make a nonempty shapes table to check
    feed = sample.copy()
    rows = [
        ["1100015", -16.743_632, 145.668_255, 10001, 1.2],
        ["1100015", -16.743_522, 145.668_394, 10002, 1.3],
    ]
    columns = [
        "shape_id",
        "shape_pt_lat",
        "shape_pt_lon",
        "shape_pt_sequence",
        "shape_dist_traveled",
    ]
    feed.shapes = pd.DataFrame(rows, columns=columns)
    assert not gkv.check_shapes(feed)

    feed1 = feed.copy()
    del feed1.shapes["shape_id"]
    assert gkv.check_shapes(feed1)

    feed1 = feed.copy()
    feed1.shapes["yo"] = 3
    assert not gkv.check_shapes(feed1)
    assert gkv.check_shapes(feed1, include_warnings=True)

    feed1 = feed.copy()
    feed1.shapes["shape_id"].iat[0] = ""
    assert gkv.check_shapes(feed1)

    for column in ["shape_pt_lon", "shape_pt_lat"]:
        feed1 = feed.copy()
        feed1.shapes[column] = 185
        assert gkv.check_shapes(feed1)

    feed1 = feed.copy()
    feed1.shapes["shape_pt_sequence"].iat[1] = feed1.shapes["shape_pt_sequence"].iat[0]
    assert gkv.check_shapes(feed1)

    feed1 = feed.copy()
    feed1.shapes["shape_dist_traveled"].iat[1] = 0
    assert gkv.check_shapes(feed1)

    feed1 = feed.copy()
    t1 = feed1.shapes.iloc[0].copy()
    t2 = feed1.shapes.iloc[1].copy()
    feed1.shapes.iloc[0] = t2
    feed1.shapes.iloc[1] = t1
    assert not gkv.check_shapes(feed1)


def test_check_stops():
    assert not gkv.check_stops(sample)

    feed = sample.copy()
    feed.stops = None
    assert gkv.check_stops(feed)

    feed = sample.copy()
    del feed.stops["stop_id"]
    assert gkv.check_stops(feed)

    feed = sample.copy()
    feed.stops["b"] = 3
    assert gkv.check_stops(feed, include_warnings=True)

    feed = sample.copy()
    feed.stops["stop_id"].iat[0] = feed.stops["stop_id"].iat[1]
    assert gkv.check_stops(feed)

    for column in ["stop_code", "stop_desc", "zone_id", "parent_station"]:
        feed = sample.copy()
        feed.stops[column] = ""
        assert gkv.check_stops(feed)

    for column in ["stop_url", "stop_timezone"]:
        feed = sample.copy()
        feed.stops[column] = "Wa wa"
        assert gkv.check_stops(feed)

    for column in [
        "stop_lon",
        "stop_lat",
        "location_type",
        "wheelchair_boarding",
    ]:
        feed = sample.copy()
        feed.stops[column] = 185
        assert gkv.check_stops(feed)

    feed = sample.copy()
    feed.stops["parent_station"] = "bingo"
    assert gkv.check_stops(feed)

    feed = sample.copy()
    feed.stops["location_type"] = 1
    feed.stops["parent_station"] = "bingo"
    assert gkv.check_stops(feed)

    feed = sample.copy()
    feed.stops["location_type"] = 0
    feed.stops["parent_station"] = feed.stops["stop_id"].iat[1]
    assert gkv.check_stops(feed)

    feed = sample.copy()
    # valid location type
    feed.stops["location_type"] = 2
    assert not gkv.check_stops(feed)
    # requires a location
    feed.stops["stop_lat"] = np.nan
    assert gkv.check_stops(feed)
    # valid location_type, does not require location
    feed.stops["location_type"] = 3
    assert not gkv.check_stops(feed)
    # valid location_type, does not require location
    feed.stops["location_type"] = 4
    assert not gkv.check_stops(feed)
    # location type 4 requires a parent station
    feed.stops["parent_station"] = np.nan
    assert gkv.check_stops(feed)
    # valid parent station for location type 4
    feed.stops["stop_lat"] = 0.0
    feed.stops["parent_station"] = feed.stops["stop_id"].iat[1]
    feed.stops["parent_station"].iat[1] = np.nan
    feed.stops["location_type"].iat[1] = 1
    assert not gkv.check_stops(feed)

    feed = sample.copy()
    feed.stops["stop_id"].iat[0] = "Flippity flew"
    assert not gkv.check_stops(feed)
    assert gkv.check_stops(feed, include_warnings=True)

    feed = sample.copy()
    feed.stop_times = None
    assert not gkv.check_stops(feed)
    assert gkv.check_stops(feed, include_warnings=True)


def test_check_stop_times():
    assert not gkv.check_stop_times(sample)

    feed = sample.copy()
    feed.stop_times = None
    assert gkv.check_stop_times(feed)

    feed = sample.copy()
    del feed.stop_times["stop_id"]
    assert gkv.check_stop_times(feed)

    feed = sample.copy()
    feed.stop_times["b"] = 3
    assert gkv.check_stop_times(feed, include_warnings=True)

    feed = sample.copy()
    feed.stop_times["trip_id"].iat[0] = "bingo"
    assert gkv.check_stop_times(feed)

    for col in ["arrival_time", "departure_time"]:
        feed = sample.copy()
        feed.stop_times[col].iat[0] = "1:0:00"
        assert gkv.check_stop_times(feed)

    feed = sample.copy()
    feed.stop_times["arrival_time"].iat[-1] = np.nan
    assert gkv.check_stop_times(feed)

    feed = sample.copy()
    feed.stop_times["stop_id"].iat[0] = "bingo"
    assert gkv.check_stop_times(feed)

    feed = sample.copy()
    feed.stop_times["stop_headsign"].iat[0] = ""
    assert gkv.check_stop_times(feed)

    feed = sample.copy()
    feed.stop_times["stop_sequence"].iat[1] = feed.stop_times["stop_sequence"].iat[0]
    assert gkv.check_stop_times(feed)

    for col in ["pickup_type", "drop_off_type"]:
        feed = sample.copy()
        feed.stop_times[col] = "bongo"
        assert gkv.check_stop_times(feed)

    feed = sample.copy()
    feed.stop_times["shape_dist_traveled"] = 1.0
    feed.stop_times["shape_dist_traveled"].iat[1] = 0.9
    assert gkv.check_stop_times(feed)

    feed = sample.copy()
    feed.stop_times["timepoint"] = 3
    assert gkv.check_stop_times(feed)

    feed = sample.copy()
    feed.stop_times["departure_time"].iat[1] = feed.stop_times["departure_time"].iat[0]
    assert not gkv.check_stop_times(feed)
    assert gkv.check_stop_times(feed, include_warnings=True)

    # Return the correct index of the missing time
    feed = sample.copy()
    # Index 2 is the first stop of the second trip
    # Index 2 is the last stop of the second trip
    feed.stop_times["arrival_time"].iat[6] = np.nan
    t1, t2 = feed.stop_times.iloc[2].copy(), feed.stop_times.iloc[6].copy()
    feed.stop_times.iloc[2], feed.stop_times.iloc[6] = t2, t1
    assert gkv.check_stop_times(feed)[0][3][0] == 2

    # Check for last stop of last trip
    # Trips are ordered by trip_id so the STBA trip_id from the sample feed
    # is last and its last stop (index 1) is the last row
    feed = sample.copy()
    feed.stop_times["arrival_time"].iat[1] = np.nan
    assert gkv.check_stop_times(feed)


def test_check_transfers():
    assert not gkv.check_transfers(sample)

    # Create transfers table
    feed = sample.copy()
    columns = [
        "from_stop_id",
        "to_stop_id",
        "transfer_type",
        "min_transfer_time",
    ]
    rows = [[feed.stops["stop_id"].iat[0], feed.stops["stop_id"].iat[1], 2, 3600]]
    feed.transfers = pd.DataFrame(rows, columns=columns)
    assert not gkv.check_transfers(feed)

    feed1 = feed.copy()
    del feed1.transfers["from_stop_id"]
    assert gkv.check_transfers(feed1)

    feed1 = feed.copy()
    feed1.transfers["yo"] = 3
    assert not gkv.check_transfers(feed1)
    assert gkv.check_transfers(feed1, include_warnings=True)

    for col in set(columns) - set(["transfer_type", "min_transfer_time"]):
        feed1 = feed.copy()
        feed1.transfers[col].iat[0] = ""
        assert gkv.check_transfers(feed1)

    for col in ["transfer_type", "min_transfer_time"]:
        feed1 = feed.copy()
        feed1.transfers[col] = -7
        assert gkv.check_transfers(feed1)


def test_check_trips():
    assert not gkv.check_trips(sample)

    feed = sample.copy()
    feed.trips = None
    assert gkv.check_trips(feed)

    feed = sample.copy()
    del feed.trips["trip_id"]
    assert gkv.check_trips(feed)

    feed = sample.copy()
    feed.trips["b"] = 3
    assert gkv.check_trips(feed, include_warnings=True)

    feed = sample.copy()
    feed.trips["trip_id"].iat[0] = feed.trips["trip_id"].iat[1]
    assert gkv.check_trips(feed)

    feed = sample.copy()
    feed.trips["route_id"] = "Hubba hubba"
    assert gkv.check_trips(feed)

    feed = sample.copy()
    feed.trips["service_id"] = "Boom boom"
    assert gkv.check_trips(feed)

    feed = sample.copy()
    feed.trips["direction_id"].iat[0] = 7
    assert gkv.check_trips(feed)

    feed = sample.copy()
    feed.trips["block_id"].iat[0] = ""
    assert gkv.check_trips(feed)

    feed = sample.copy()
    feed.trips["shape_id"].iat[0] = "Hello"
    assert gkv.check_trips(feed)

    feed = sample.copy()
    feed.trips["wheelchair_accessible"] = ""
    assert gkv.check_trips(feed)

    feed = sample.copy()
    tid = feed.trips["trip_id"].iat[0]
    feed.stop_times = feed.stop_times[feed.stop_times["trip_id"] != tid].copy()
    assert not gkv.check_trips(feed)
    assert gkv.check_trips(feed, include_warnings=True)

    feed = sample.copy()
    feed.stop_times = None
    assert not gkv.check_trips(feed)
    assert gkv.check_trips(feed, include_warnings=True)


def test_validate():
    assert not gkv.validate(sample, as_df=False, include_warnings=False)



"""
Functions about cleaning feeds.
"""

from __future__ import annotations
from typing import TYPE_CHECKING

import pandas as pd
import numpy as np

from . import constants as cs
from . import helpers as hp

# Help mypy but avoid circular imports
if TYPE_CHECKING:
    from .feed import Feed


def clean_column_names(df: pd.DataFrame) -> pd.DataFrame:
    """
    Strip the whitespace from all column names in the given DataFrame
    and return the result.
    """
    f = df.copy()
    f.columns = [col.strip() for col in f.columns]
    return f


def clean_ids(feed: "Feed") -> "Feed":
    """
    In the given Feed, strip whitespace from all string IDs and
    then replace every remaining whitespace chunk with an underscore.
    Return the resulting Feed.
    """
    # Alter feed inputs only, and build a new feed from them.
    # The derived feed attributes, such as feed.trips_i,
    # will be automatically handled when creating the new feed.
    feed = feed.copy()

    for table in cs.GTFS_REF["table"].unique():
        f = getattr(feed, table)
        if f is None:
            continue
        for column in cs.GTFS_REF.loc[cs.GTFS_REF["table"] == table, "column"]:
            if column in f.columns and column.endswith("_id"):
                try:
                    f[column] = (
                        f[column].str.strip().str.replace(r"\s+", "_", regex=True)
                    )
                    setattr(feed, table, f)
                except AttributeError:
                    # Column is not of string type
                    continue

    return feed


def extend_id(feed: "Feed", id_col: str, extension: str, *, prefix=True) -> "Feed":
    """
    Add a prefix (if ``prefix``) or a suffix (otherwise) to all values of column
    ``id_col`` across all tables of this Feed.
    This can be helpful when preparing to merge multiple GTFS feeds with colliding
    route IDs, say.

    Raises a ValueError if ``id_col`` values can't have strings added to them,
    e.g. if ``id_col`` is 'direction_id'.
    """
    feed = feed.copy()

    for table in cs.GTFS_REF.loc[lambda x: x["column"] == id_col, "table"].unique():
        t = getattr(feed, table)
        if t is None:
            continue
        try:
            t[id_col] = extension + t[id_col] if prefix else t[id_col] + extension
            setattr(feed, table, t)
        except Exception as e:
            raise ValueError(e)

    return feed


def clean_times(feed: "Feed") -> "Feed":
    """
    In the given Feed, convert H:MM:SS time strings to HH:MM:SS time
    strings to make sorting by time work as expected.
    Return the resulting Feed.
    """

    def reformat(t):
        if pd.isna(t):
            return t
        t = t.strip()
        if len(t) == 7:
            t = "0" + t
        return t

    feed = feed.copy()
    tables_and_columns = [
        ("stop_times", ["arrival_time", "departure_time"]),
        ("frequencies", ["start_time", "end_time"]),
    ]
    for table, columns in tables_and_columns:
        f = getattr(feed, table)
        if f is not None:
            f[columns] = f[columns].map(reformat)
        setattr(feed, table, f)

    return feed


def drop_zombies(feed: "Feed") -> "Feed":
    """
    In the given Feed, do the following in order and return the resulting Feed.

    1. Drop stops of location type 0 or NaN with no stop times.
    2. Remove undefined parent stations from the ``parent_station`` column.
    3. Drop trips with no stop times.
    4. Drop shapes with no trips.
    5. Drop routes with no trips.
    6. Drop services with no trips.

    """
    feed = feed.copy()

    f = feed.stops.copy()
    ids = feed.stop_times.stop_id.unique()
    cond = f.stop_id.isin(ids)
    if "location_type" in f.columns:
        cond |= ~f.location_type.isin([0, np.nan])
    feed.stops = f[cond].copy()

    # Remove undefined parent stations from the ``parent_station`` column
    if "parent_station" in feed.stops.columns:
        f = feed.stops.copy()
        ids = f.stop_id.unique()
        f["parent_station"] = f.parent_station.map(lambda x: x if x in ids else np.nan)
        feed.stops = f

    # Drop trips with no stop times
    ids = feed.stop_times["trip_id"].unique()
    f = feed.trips
    feed.trips = f[f["trip_id"].isin(ids)]

    # Drop shapes with no trips
    ids = feed.trips["shape_id"].unique()
    f = feed.shapes
    if f is not None:
        feed.shapes = f[f["shape_id"].isin(ids)]

    # Drop routes with no trips
    ids = feed.trips["route_id"].unique()
    f = feed.routes
    feed.routes = f[f["route_id"].isin(ids)]

    # Drop services with no trips
    ids = feed.trips["service_id"].unique()
    if feed.calendar is not None:
        f = feed.calendar
        feed.calendar = f[f["service_id"].isin(ids)]
    if feed.calendar_dates is not None:
        f = feed.calendar_dates
        feed.calendar_dates = f[f["service_id"].isin(ids)]

    return feed


def clean_route_short_names(feed: "Feed") -> "Feed":
    """
    In ``feed.routes``, assign 'n/a' to missing route short names and
    strip whitespace from route short names.
    Then disambiguate each route short name that is duplicated by
    appending '-' and its route ID.
    Return the resulting Feed.
    """
    feed = feed.copy()
    r = feed.routes
    if r is None:
        return feed

    # Fill NaNs and strip whitespace
    r["route_short_name"] = r["route_short_name"].fillna("n/a").str.strip()

    # Disambiguate
    def disambiguate(row):
        rsn, rid = row
        return rsn + "-" + rid

    r["dup"] = r["route_short_name"].duplicated(keep=False)
    r.loc[r["dup"], "route_short_name"] = r.loc[
        r["dup"], ["route_short_name", "route_id"]
    ].apply(disambiguate, axis=1)
    del r["dup"]

    feed.routes = r
    return feed


def build_aggregate_routes_dict(
    routes: pd.DataFrame, by: str = "route_short_name", route_id_prefix: str = "route_"
) -> dict[str, str]:
    """
    Given a DataFrame of routes, group the routes by route short name, say,
    and assign new route IDs using the given prefix.
    Return a dictionary of the form <old route ID> -> <new route ID>.
    Helper function for :func:`aggregate_routes`.

    More specifically, group ``routes`` by the ``by`` column, and for each group make
    one new route ID for all the old route IDs in that group based on the given
    ``route_id_prefix`` string and a running count, e.g. ``'route_013'``.
    """
    if by not in routes.columns:
        raise ValueError(f"Column {by} not in routes.")

    # Create new route IDs
    n = routes.groupby(by).ngroups
    nids = hp.make_ids(n, route_id_prefix)
    nid_by_oid = dict()
    i = 0
    for col, group in routes.groupby(by):
        d = {oid: nids[i] for oid in group.route_id.values}
        nid_by_oid.update(d)
        i += 1

    return nid_by_oid


def aggregate_routes(
    feed: "Feed", by: str = "route_short_name", route_id_prefix: str = "route_"
) -> "Feed":
    """
    Aggregate routes by route short name, say, and assign new route IDs using the
    given prefix.

    More specifically, create new route IDs with the function
    :func:`build_aggregate_routes_dict` and the parameters ``by`` and
    ``route_id_prefix`` and update the old route IDs to the new ones in all the relevant
    Feed tables.
    Return the resulting Feed.
    """
    feed = feed.copy()

    # Make new route IDs
    routes = feed.routes
    nid_by_oid = build_aggregate_routes_dict(routes, by, route_id_prefix)

    # Update route IDs in routes
    routes["route_id"] = routes.route_id.map(lambda x: nid_by_oid[x])
    routes = routes.groupby(by).first().reset_index()
    feed.routes = routes

    # Update route IDs in trips
    trips = feed.trips
    trips["route_id"] = trips.route_id.map(lambda x: nid_by_oid[x])
    feed.trips = trips

    # Update route IDs of fare rules
    if feed.fare_rules is not None and "route_id" in feed.fare_rules.columns:
        fr = feed.fare_rules
        fr["route_id"] = fr.route_id.map(lambda x: nid_by_oid[x])
        feed.fare_rules = fr

    return feed


def build_aggregate_stops_dict(
    stops: pd.DataFrame, by: str = "stop_code", stop_id_prefix: str = "stop_"
) -> dict[str, str]:
    """
    Given a DataFrame of stops, group the stops by stop code, say,
    and assign new stop IDs using the given prefix.
    Return a dictionary of the form <old stop ID> -> <new stop ID>.
    Helper function for :func:`aggregate_stops`.

    More specifically, group ``stops`` by the ``by`` column, and for each group make
    one new stop ID for all the old stops IDs in that group based on the given
    ``stop_id_prefix`` string and a running count, e.g. ``'stop_013'``.
    """
    if by not in stops.columns:
        raise ValueError(f"Column {by} not in stops.")

    # Create new stop IDs
    n = stops.groupby(by).ngroups
    nids = hp.make_ids(n, stop_id_prefix)
    nid_by_oid = dict()
    i = 0
    for col, group in stops.groupby(by):
        d = {oid: nids[i] for oid in group.stop_id.values}
        nid_by_oid.update(d)
        i += 1

    return nid_by_oid


def aggregate_stops(
    feed: "Feed", by: str = "stop_code", stop_id_prefix: str = "stop_"
) -> "Feed":
    """
    Aggregate stops by stop code, say, and assign new stop IDs using the
    given prefix.

    More specifically, create new stop IDs with the function
    :func:`build_aggregate_stops_dict` and the parameters ``by`` and
    ``stop_id_prefix`` and update the old stop IDs to the new ones in all the relevant
    Feed tables.
    Return the resulting Feed.
    """
    feed = feed.copy()

    # Make new stop ID by old stop ID dict
    stops = feed.stops
    nid_by_oid = build_aggregate_stops_dict(stops, by, stop_id_prefix)

    # Apply dict
    stops["stop_id"] = stops.stop_id.map(nid_by_oid)
    if "parent_station" in stops:
        stops["parent_station"] = stops.parent_station.map(nid_by_oid)

    stops = stops.groupby(by).first().reset_index()
    feed.stops = stops

    # Update stop IDs of stop times
    stop_times = feed.stop_times
    stop_times["stop_id"] = stop_times.stop_id.map(lambda x: nid_by_oid[x])
    feed.stop_times = stop_times

    # Update route IDs of transfers
    if feed.transfers is not None:
        transfers = feed.transfers
        transfers["to_stop_id"] = transfers.to_stop_id.map(lambda x: nid_by_oid[x])
        transfers["from_stop_id"] = transfers.from_stop_id.map(lambda x: nid_by_oid[x])
        feed.transfers = transfers

    return feed


def clean(feed: "Feed") -> "Feed":
    """
    Apply the following functions to the given Feed in order and return the resulting
    Feed.

    #. :func:`clean_ids`
    #. :func:`clean_times`
    #. :func:`clean_route_short_names`
    #. :func:`drop_zombies`

    """
    feed = feed.copy()
    ops = ["clean_ids", "clean_times", "clean_route_short_names", "drop_zombies"]
    for op in ops:
        feed = globals()[op](feed)

    return feed


def drop_invalid_columns(feed: "Feed") -> "Feed":
    """
    Drop all DataFrame columns of the given Feed that are not
    listed in the GTFS.
    Return the resulting Feed.
    """
    feed = feed.copy()
    for table, group in cs.GTFS_REF.groupby("table"):
        f = getattr(feed, table)
        if f is None:
            continue
        valid_columns = group["column"].values
        for col in f.columns:
            if col not in valid_columns:
                print(f"{table}: dropping invalid column {col}")
                del f[col]
        setattr(feed, table, f)

    return feed
